from .utils import ModdelingOutput, get_inference_system, get_offload_system
from GenZ.unit import Unit
from GenZ.operators import *

from GenZ.operator_base import op_type_dicts
from GenZ.system import System
import pandas as pd
from GenZ.analyse_model import *
import warnings
from GenZ.LLM_inference import decode_moddeling, prefill_moddeling
from paretoset import paretoset
import itertools
from GenZ.Models import get_configs, create_inference_moe_prefill_layer, create_inference_moe_decode_layer

unit = Unit()

def factors(n):
    return [x for tup in ([i, n//i]
                for i in range(1, int(n**0.5)+1) if n % i == 0) for x in tup]

def get_various_parallization(model='llama2_7b', total_nodes=8):
    model_config = get_configs(model)

    if total_nodes == 1:
        return {(1,1)}
    elif total_nodes < 1:
        raise ValueError(f'Num of Nodes:{total_nodes} should be >= 1')

    H = model_config.num_attention_heads
    num_layers = model_config.num_decoder_layers

    TP_parallelism = np.sort(factors(H))[::-1]
    PP_parallelism = np.sort(factors(num_layers))[::-1]

    parallelism_combinations = set()

    for TP, PP in itertools.product(TP_parallelism, PP_parallelism):
        if TP * PP < total_nodes and TP*PP >= total_nodes//2:
            parallelism_combinations.add((TP, PP))
    return parallelism_combinations

def get_best_parallization_strategy(
        stage='decode', model='llama2_7b', total_nodes=8, batch_size = 1, beam_size = 1,
        input_tokens = 2000, output_tokens = 256,
        system_name = {'Flops': 200, 'Memory_size': 32, 'Memory_BW': 1000, 'ICN': 300 , 'real_values':True},
        bits='bf16', debug=False
        ):

    parallelism_combinations = get_various_parallization(model=model, total_nodes=total_nodes)
    if debug:
        print(f'For model:{model}, number cores:{total_nodes}, system:{system_name}, \n The parallelism combinations are {parallelism_combinations} ')

    data = []
    for TP,PP in parallelism_combinations:
        if PP <= batch_size:
            micro_batch_size = (batch_size//PP)
            if stage == 'prefill':
                prefill_outputs = prefill_moddeling(model = model, batch_size = micro_batch_size,
                                        input_tokens = input_tokens,
                                        system_name = system_name,
                                        bits=bits,
                                        tensor_parallel = TP, pipeline_parallel = PP, debug=debug)
                data.append([micro_batch_size, TP, PP , prefill_outputs['Latency'], prefill_outputs['Throughput']] + prefill_outputs['Runtime_breakdown'])
            elif stage == 'decode':
                decode_outputs = decode_moddeling(model = model, batch_size = micro_batch_size, Bb = beam_size ,
                                    input_tokens = input_tokens, output_tokens = output_tokens,
                                    system_name = system_name,
                                    bits=bits,
                                    tensor_parallel = TP, pipeline_parallel =PP, debug=debug)
                data.append([micro_batch_size, TP, PP,  decode_outputs['Latency'], decode_outputs['Throughput']] + decode_outputs['Runtime_breakdown'])
            else:
                raise ValueError('Stage should be prefill or decode')

    data_df = pd.DataFrame(data, columns = ['micro batch', 'TP', 'PP', 'Latency(ms)', 'Tokens/s', 'GEMM time', 'SA time', 'Comm. time'])
    if debug:
        display(data_df)
    return data_df.sort_values(by='Tokens/s', ascending=False).head(1)

def get_pareto_optimal_performance(
        stage='decode', model='llama2_7b', total_nodes=8, batch_list = 1, beam_size = 1,
        input_tokens = 2000, output_tokens = 256,
        system_name = {'Flops': 200, 'Memory_size': 32, 'Memory_BW': 1000, 'ICN': 300 , 'real_values':True},
        bits='bf16', debug=False
        ):

    parallelism_combinations = get_various_parallization(model=model, total_nodes=total_nodes)

    if debug:
        print(f'For model:{model}, number cores:{total_nodes}, system:{system_name}, \n The parallelism combinatations are {parallelism_combinations} ')
    data = []
    if isinstance(batch_list, int):
        batch_list = [batch_list]
    for batch_size in batch_list:
        for TP,PP in parallelism_combinations:
            if PP <= batch_size:
                micro_batch_size = (batch_size//PP)
                if stage == 'prefill':
                    prefill_outputs = prefill_moddeling(model = model, batch_size = micro_batch_size,
                                            input_tokens = input_tokens,
                                            system_name = system_name,
                                            bits=bits,
                                            tensor_parallel = TP, pipeline_parallel = PP, debug=False)
                    data.append([batch_size, micro_batch_size, TP, PP , prefill_outputs['Latency'], prefill_outputs['Throughput']] + prefill_outputs['Runtime_breakdown'])
                elif stage == 'decode':
                    decode_outputs = decode_moddeling(model = model, batch_size = micro_batch_size, Bb = beam_size ,
                                        input_tokens = input_tokens, output_tokens = output_tokens, 
                                        system_name = system_name,
                                        bits=bits,
                                        tensor_parallel = TP, pipeline_parallel =PP, debug=False)
                    data.append([batch_size, micro_batch_size, TP, PP,  decode_outputs['Latency'], decode_outputs['Throughput']] + decode_outputs['Runtime_breakdown'])
                else:
                    raise ValueError('Stage should be prefill or decode')

    data_df = pd.DataFrame(data, columns = ['batch', 'micro batch', 'TP', 'PP', 'Latency(ms)', 'Tokens/s', 'GEMM time', 'SA time', 'Comm. time'])
    datapoints = data_df[['Latency(ms)','Tokens/s']]

    ##  We want a pareto optimal frontier with minimum latency and maximum Throughput.
    mask = paretoset(datapoints, sense=["min", "max"])
    return data_df[mask]


from .LLM_inference import (
    ModdelingOutput,
    get_inference_system,
    get_offload_system,
    decode_moddeling,
    prefill_moddeling,
    get_minimum_system_size,
    factors,
    get_various_parallization,
    get_best_parallization_strategy,
    get_pareto_optimal_performance,
)
from .system import System
from .unit import Unit
from .analyse_model import get_model_df, get_summary_table, simplify_df
from .collective_times import get_AR_time, get_message_pass_time
from .Models import (
    ModelConfig,
    get_configs,
    create_inference_moe_prefill_layer,
    create_inference_moe_decode_layer,
    create_inference_mamba_prefix_model,
    create_inference_mamba_decode_model,
    create_full_prefill_model,
    create_full_decode_model,
)
from .parallelism import ParallelismConfig


import numpy as np
import math
from GenZ.unit import Unit
class System(object):
    compute_multiplier = {'int8': 0.5, 'bf16': 1, 'f32': 2, 'int4': 0.25, 'int2':0.125, 'fp8': 0.5, 'fp4': 0.25}
    mem_multiplier = {'int8': 1, 'bf16': 2, 'f32': 4, 'int4':0.5, 'int2':0.25, 'fp8':1, 'fp4':0.5}
    def __init__(self, unit=None,
                flops=123, mxu_shape=None,
                onchip_mem_bw=18000, on_chip_mem_size=float('Inf'),
                offchip_mem_bw=900, off_chip_mem_size=float('Inf'),
                external_mem_bw=100,
                frequency=940, bits='bf16',
                compute_efficiency=1, memory_efficiency=1, comm_efficiency=1,
                interchip_link_bw = 25, num_nodes = 1, interchip_link_latency=1.9,
                collective_strategy='GenZ',    # GenZ or ASTRA-SIM
                topology='FullyConnected',
                parallelism_heirarchy = "TP{1}_EP{1}_PP{1}",
                network_config = None
                ):

        if unit is None:
            self.unit = Unit()
        else:
            self.unit = unit

        self.flops = self.unit.unit_to_raw(flops, type='C')
        self.op_per_sec = self.flops/2

        self.frequency = self.unit.unit_to_raw(frequency, type='F')
        self.onchip_mem_bw = self.unit.unit_to_raw(onchip_mem_bw, type='BW')
        self.offchip_mem_bw = self.unit.unit_to_raw(offchip_mem_bw, type='BW')
        self.interchip_link_bw = self.unit.unit_to_raw(interchip_link_bw, type='BW')
        self.interchip_link_latency = interchip_link_latency * 1e-6     ## us
        self.external_mem_bw = self.unit.unit_to_raw(external_mem_bw, type='BW')
        self.on_chip_mem_size = self.unit.unit_to_raw(on_chip_mem_size, type='M')
        self.on_chip_mem_left_size = self.unit.unit_to_raw(on_chip_mem_size, type='M')
        self.off_chip_mem_size = self.unit.unit_to_raw(off_chip_mem_size, type='M')
        self.compute_efficiency = compute_efficiency
        self.memory_efficiency = memory_efficiency
        self.comm_efficiency = comm_efficiency
        self.mxu_shape = mxu_shape
        
        self.collective_strategy = collective_strategy
        assert self.collective_strategy in ['GenZ', 'ASTRA-SIM'], "Invalid collective_strategy. Must be one of: GenZ, ASTRA-SIM"
        self.num_nodes = num_nodes
        self.topology = topology
        self.bits = bits
        self.parallelism_heirarchy = parallelism_heirarchy   ## TP{1}_EP{1}_PP{1}
        self.network_config = network_config

    def __str__(self):
        unit = Unit()
        a = f"Accelerator OPS: {unit.raw_to_unit(self.flops,type='C')} TOPS , Freq = {unit.raw_to_unit(self.frequency,type='F')} GHz, Num Nodes = {self.num_nodes} \n"
        b = f"On-Chip mem size: {unit.raw_to_unit(self.on_chip_mem_size, type='M')} MB , Off-chip mem size:{unit.raw_to_unit(self.off_chip_mem_size, type='M')} MB\n"
        c = f"On-Chip mem BW: {unit.raw_to_unit(self.onchip_mem_bw, type='BW')} GB/s , Off-chip mem BW:{unit.raw_to_unit(self.offchip_mem_bw, type='BW')} GB/s, External-mem BW:{unit.raw_to_unit(self.external_mem_bw, type='BW')} GB/s,\n"
        return a+b+c

    def get_params(self):
        unit = Unit()
        a = f"Accelerator OPS: {unit.raw_to_unit(self.flops,type='C')} TOPS , Freq = {unit.raw_to_unit(self.frequency,type='F')} GHz, Num Nodes = {self.num_nodes}"
        b = f" Off-chip mem size:{unit.raw_to_unit(self.off_chip_mem_size, type='M')/1024} GB "
        c = f" Off-chip mem BW:{unit.raw_to_unit(self.offchip_mem_bw, type='BW')} GB/s, External-mem BW:{unit.raw_to_unit(self.external_mem_bw, type='BW')} GB/s"
        return a+b+c


    def set_onchip_mem_bw(self,onchip_mem_bw):
        self.onchip_mem_bw = self.unit.unit_to_raw(onchip_mem_bw, type='BW')

    def set_offchip_mem_bw(self,offchip_mem_bw):
        self.offchip_mem_bw = self.unit.unit_to_raw(offchip_mem_bw, type='BW')

    def get_offchip_mem_bw(self):
        return self.unit.raw_to_unit(self.offchip_mem_bw,type='BW')

    def get_external_mem_bw(self):
        return self.unit.raw_to_unit(self.external_mem_bw,type='BW')

    def get_interchip_link_bw(self):
        return self.unit.raw_to_unit(self.interchip_link_bw,type='BW')

    def get_off_chip_mem_size(self):
        return self.unit.raw_to_unit(self.off_chip_mem_size,type='M')


    def claim_onchip_mem(self, data_sz):
        if data_sz > self.on_chip_mem_left_size:
            raise ValueError(f'Not enough on-chip memory: Need {data_sz}, only has {self.on_chip_mem_size}')
        self.on_chip_mem_left_size -= data_sz
        return self.on_chip_mem_left_size

    def release_onchip_mem(self, data_sz):
        self.on_chip_mem_left_size = max(self.on_chip_mem_size, data_sz + self.on_chip_mem_left_size)
        return self.on_chip_mem_left_size

    def get_bit_multiplier(self, type='C', data='a'):
        if self.bits == 'special':
            if data == 'w':
                return 3/8
            else:
                return 2
        if type == 'C':
            return self.compute_multiplier[self.bits]
        elif type == 'M':
            return self.mem_multiplier[self.bits]


from GenZ.Models import ModelConfig, ResidencyInfo, OpType, CollectiveType
from GenZ.parallelism import ParallelismConfig
from math import ceil

def mha_flash_attention_prefill(model_config:ModelConfig, parallelism_config:ParallelismConfig, input_sequence_length:int):
    H = model_config.num_attention_heads
    Hkv = model_config.num_key_value_heads
    D = model_config.hidden_size
    Dq = model_config.head_dim

    tp = parallelism_config.tensor_parallel * parallelism_config.expert_parallel
    sp = parallelism_config.sequence_parallel
    per_node_H = max(ceil(H / tp), 1)
    per_node_Hkv = max(ceil(Hkv / tp), 1)

    ## [Batch/dp, Seq/sp, Dmodel] * [2, Dmodel, Dq, Hkv/tp] + [Dmodel, Dq, Head/tp]= [Batch/dp, Seq/sp, 3, Dq, Head/tp]
    QKV =           [["QKV", (per_node_H*Dq + 2*per_node_Hkv*Dq), input_sequence_length//sp, D, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]

    ## [Batch/dp, Seq, Dq, Head/tp] * [Batch/dp, Seq/sp, Dq, Head/tp] = [Batch/dp, Seq, Seq/sp, Head/tp]
    logit =         [["Logit",per_node_H, input_sequence_length, input_sequence_length//sp, Dq, per_node_Hkv, ResidencyInfo.C_onchip, OpType.Logit]]

    ## [Batch/dp, Seq, Seq/sp, Head/tp] * [Batch/dp, Seq/sp, Dq, Head/tp] = [Batch/dp, Seq, Dq, Head/tp]
    attend =        [["Attend",per_node_H, input_sequence_length, input_sequence_length//sp, Dq, per_node_Hkv, ResidencyInfo.A_onchip, OpType.Attend]]

    ## [Batch/dp, Seq, Dq, Head/tp] * [Dq, Head/tp,  Dmodel] = [Batch/dp, Seq, Dmodel]
    output =        [["Out Proj", D, input_sequence_length//sp, (per_node_H) * Dq, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]

    if tp > 1:
        sync =          [["MHA AR", input_sequence_length//sp, D, 1, 1, tp, CollectiveType.AllReduce, OpType.Sync]]
    else:
        sync = []
    return QKV + logit + attend + output + sync

def mha_flash_attention_decode(model_config:ModelConfig, parallelism_config:ParallelismConfig, input_sequence_length:int, output_gen_tokens:int):
    H = model_config.num_attention_heads
    Hkv = model_config.num_key_value_heads
    D = model_config.hidden_size
    Dq = model_config.head_dim

    tp = parallelism_config.tensor_parallel * parallelism_config.expert_parallel
    sp = parallelism_config.sequence_parallel
    dp = parallelism_config.data_parallel

    per_node_H = max(ceil(H / tp), 1)
    per_node_Hkv = max(ceil(Hkv / tp), 1)

    query =         [["QKV", (per_node_H*Dq + 2*per_node_Hkv*Dq), 1, D, 1, 1, ResidencyInfo.AC_onchip, OpType.GEMM]]
    logit_pre =     [["Logit Pre",per_node_H, 1, input_sequence_length//sp, Dq, per_node_Hkv, ResidencyInfo.AC_onchip, OpType.Logit_BM_PREFILL]]
    attend_pre =    [["Attend Pre",per_node_H, 1, input_sequence_length//sp, Dq, per_node_Hkv, ResidencyInfo.AC_onchip, OpType.Attend_BM_PREFILL]]
    logit_suf =     [["Logit Suf",per_node_H, 1, output_gen_tokens, Dq, per_node_Hkv, ResidencyInfo.AC_onchip, OpType.Logit]]
    attend_suf =    [["Attend Suf",per_node_H, 1, output_gen_tokens, Dq, per_node_Hkv, ResidencyInfo.AC_onchip, OpType.Attend]]
    output =        [["Out Proj",D, 1, (per_node_H) * Dq, 1, 1, ResidencyInfo.AC_onchip, OpType.GEMM]]
    if tp > 1:
        sync =          [["MHA AR",1, D, 1, 1, tp, CollectiveType.AllReduce, OpType.Sync]]
    else:
        sync = []

    return query + logit_pre + logit_suf + attend_pre + attend_suf + output + sync




import os
import subprocess
import yaml
from GenZ.system import System
from GenZ.unit import Unit
import numpy as np

from .fix_chakra_traces import convert_chakra_file
import re

txt_file_path = "/tmp/genz/chakra/txt_file.txt"
et_output_path = "/tmp/genz/chakra/et/"
et_cleaned_output_path = "/tmp/genz/chakra/et_cleaned"

SCRIPT_DIR=os.path.dirname(os.path.realpath(__file__))
run_file = os.path.join(SCRIPT_DIR, "run.sh")
ASTRA_SIM_OUTPUT_PATH = os.path.join(SCRIPT_DIR, "astra_output.txt") 

def divide_npus_count(network_config, parallelism_sizes):
    result = []
    dims = []
    npus_count = network_config["npus_count"].copy()
    dim_index = np.arange(len(npus_count)).tolist()
    for size in parallelism_sizes:
        current_nodes = []
        current_dims = []
        while np.prod(current_nodes) < size:
            last_dim_nodes = npus_count.pop(0)
            index = dim_index.pop(0)
            if last_dim_nodes * max(1, np.prod(current_nodes)) <= size:
                current_nodes.append(last_dim_nodes)
                current_dims.append(index)
            else:
                dim_used = size/max(1, np.prod(current_nodes))
                current_nodes.append(dim_used)
                current_dims.append(index)
                last_dim_nodes /= dim_used
                npus_count.insert(0, last_dim_nodes)
                dim_index.insert(0, index)
        result.append(current_nodes)
        dims.append(current_dims)
    return result, dims

def get_network_config(network_config:dict, parallelism_heirarchy:str, parallelism:str) -> dict:
    """
    network_config: dict: A dictionary with the following keys:
            "topology":   ## List of topology (“Ring”, “FullyConnected”, or “Switch”)
            "npus_count": ## List of number of npus per node
            "bandwidth":  ## List of Link Bw in GB/s
            "latency":    ## Link expects latency in ns
    parallelism_heirarchy: str: A string with the following format: Ex: "TP{x}_EP{y}_PP{z}"
        In the above example, TP is among the closest nodes, then EP and finally PP accross the last dimension.
    """ 
    assert type(network_config) == dict, "network_config must be a dictionary"
    
    pattern = r'\{(\d+)\}'
    parallelism_sizes = re.findall(pattern, parallelism_heirarchy)
    assert np.prod(network_config["npus_count"]) == np.prod([int(match) for match in parallelism_sizes]), f"Prof of npus_count:{np.prod(network_config['npus_count'])} should be equal to num_nodes:{np.prod([int(match) for match in parallelism_sizes])}"

    assert parallelism in parallelism_heirarchy, "parallelism should be present in parallelism_heirarchy"

    # find the parallelism in parallelism_heirarchy, it would be distributed by '_', find the position. 
    # Ex1:parallelism_heirarchy = 'TP{2}_EP{4}_PP{2}' , parallelism = EP
    # output = 1
    # Ex2:parallelism_heirarchy = 'TP{2}_EP{4}_PP{2}' , parallelism = TP
    # output = 0
    parallelism_position = re.sub(pattern,'',parallelism_heirarchy ).split('_').index(parallelism)
    
    dims, indexes = divide_npus_count(network_config, [int(match) for match in parallelism_sizes])

    parallelism_index = indexes[parallelism_position]
    network_config_for_parallelism = {}
    for key, value in network_config.items():
        if key == "npus_count":
            network_config_for_parallelism[key] = [int(x) for x in dims[parallelism_position]]
        else:
            network_config_for_parallelism[key] = [value[i] for i in parallelism_index]

    network_config_for_parallelism["dimensions-count"] = len(dims[parallelism_position])
    return network_config_for_parallelism

import json

def replace_collective_implementation(file_path, new_value):
    with open(file_path, 'r') as file:
        data = json.load(file)
    
    data["all-reduce-implementation"] = new_value
    data["all-to-all-implementation"] = new_value
    data["all-gather-implementation"] = new_value
    data["reduce-scatter-implementation"] = new_value
    
    with open(file_path, 'w') as file:
        json.dump(data, file, indent=4)

# Example usage


def get_astrasim_collective_time(collective_size, collective_type, system:System=None,
                                network_config=None) -> dict:
    """
    collective_type: str: Should be one of the following: "ALLREDUCE"/"ALLTOALL"/"ALLGATHER"/"REDUCESCATTER"
    collective_size: int: Size of the collective operation in Bytes
    
    Returns: dict: A dictionary with the key as the system id and the value is latency in ns
    
    """
    assert collective_type in ["ALLREDUCE", "ALLTOALL", "ALLGATHER", "REDUCESCATTER"], \
        "Invalid collective_type. Must be one of: ALLREDUCE, ALLTOALL, ALLGATHER, REDUCESCATTER"
    # Step 1: Create the text file
    os.makedirs(os.path.dirname(txt_file_path), exist_ok=True)
    with open(txt_file_path, "w+") as txt_file:
        txt_file.write("MICRO\n1\nDUMMYNAME -1 5 NONE 0 5 NONE 0 5 {} {} 5\n".format(collective_type, collective_size))
    
    # Step 2: Generate ET trace
    os.makedirs(os.path.dirname(et_output_path), exist_ok=True)

    if network_config is not None:
        nodes = int(np.prod(network_config["npus_count"]))
    else:
        nodes = system.num_nodes
    # Wait for the subprocess to complete
    subprocess.run([
        "python", "-m", "chakra.src.converter.converter", "Text",
        "--input", txt_file_path,
        "--output", et_output_path + "collective_traces",
        "--num-npus", str(nodes),
        "--num-passes", "1"
    ], check=True)
    # Step 3: Clean up the generated traces
    os.makedirs(et_cleaned_output_path, exist_ok=True)
    files = os.listdir(et_output_path)
    for file in files:
        if not file.endswith(".et"):
            continue
        input_file = os.path.join(et_output_path, file)
        output_file = os.path.join(et_cleaned_output_path, file)
        convert_chakra_file(input_file, output_file)
    
    # Step 4: Create network.yml
    if network_config is None:
        assert system.topology in ["Ring", "FullyConnected", "Switch"], \
            "Invalid collective_type. Must be one of: Ring, FullyConnected, Switch"
        network_config = {
            "topology": [system.topology],    #(“Ring”, “FullyConnected”, or “Switch”)
            "npus_count": [system.num_nodes],
            "bandwidth": [Unit().raw_to_unit(system.interchip_link_bw, type="BW")],
            "latency": [system.interchip_link_latency*1e9],## AStrasim expects latency in ns 
        }
    else:
        assert type(network_config) == dict, "network_config must be a dictionary"

    network_yml_path = SCRIPT_DIR+"/network.yml"
    with open(network_yml_path, "w+") as network_yml_file:
        yaml.dump(network_config, network_yml_file)
    
    # Step 5: Replace the system implementation
    topology_to_algorithm = {
        "Ring": "ring",
        "FullyConnected": "direct",
        "Switch": "halvingDoubling"
    }
    collective_impl = [topology_to_algorithm[i] for i in network_config['topology']]
    replace_collective_implementation('/home/abhimanyu/synergy3/work/GenZ-LLM-Analyzer/GenZ/Astra_sim/system.json', collective_impl) 
    # Step 6: Run astra-sim
    print(subprocess.run(f"bash {run_file}>{ASTRA_SIM_OUTPUT_PATH}", shell=True, check=True, stderr=subprocess.PIPE).stdout)

    def get_cycles_count(file_path):
        cycles_count = {}
        with open(file_path, 'r') as file:
            for line in file:
                if 'finished' in line:
                    parts = line.split(',')
                    sys_id = parts[0].split('[')[1].split(']')[0]
                    cycles = int(parts[1].split()[0])
                    cycles_count[sys_id] = cycles
        return cycles_count
    
    cycles_count = get_cycles_count(ASTRA_SIM_OUTPUT_PATH)
    return cycles_count


from ..default_models import ModelConfig, get_all_model_configs

#### OPT Models ####

# https://huggingface.co/openai-community/gpt2/blob/main/config.json
gpt2_config = ModelConfig(model='openai/gpt2',
    hidden_size=768, num_attention_heads=12, num_ffi = 1,
    intermediate_size=4*768, num_decoder_layers=12,
    vocab_size=50257, hidden_act="gelu_new", max_model_len=1024
)

# https://huggingface.co/facebook/opt-125m/blob/main/config.json
opt_125m_config = ModelConfig(model='facebook/opt-125M',
    hidden_size=768, num_attention_heads=12, num_ffi = 1,
    intermediate_size=4*768, num_decoder_layers=12,
    vocab_size=50272, hidden_act="relu", max_model_len=2*1024
)

# https://huggingface.co/facebook/opt-350m/blob/main/config.json
opt_350m_config = ModelConfig(model='facebook/OPT-350M',
    hidden_size=1024, num_attention_heads=16, num_ffi = 1,
    intermediate_size=4*1024, num_decoder_layers=24,
    vocab_size=50272, hidden_act="relu", max_model_len=2*1024
)

# https://huggingface.co/facebook/opt-1.3b/blob/main/config.json
opt_1b_config = ModelConfig(model='facebook/OPT-1.3b',
    hidden_size=2048, num_attention_heads=32, num_ffi = 1,
    intermediate_size=4*2048, num_decoder_layers=24,
    max_model_len=2*1024, vocab_size=50272, hidden_act="relu",
)

# https://huggingface.co/facebook/opt-6.7b/blob/main/config.json
opt_7b_config = ModelConfig(model='facebook/opt-6.7b',
    hidden_size=4096, num_attention_heads=32, num_ffi = 1,
    intermediate_size=4*4096, num_decoder_layers=32,
    max_model_len=2*1024, vocab_size=50272, hidden_act="relu",
)

# https://huggingface.co/facebook/opt-13b/blob/main/config.json
opt_13b_config = ModelConfig(model='facebook/OPT-13B',
    hidden_size=5140, num_attention_heads=40, num_ffi = 1,
    intermediate_size=4*5140, num_decoder_layers=40,
    max_model_len=2*1024, vocab_size=50272, hidden_act="relu",
)

# https://huggingface.co/facebook/opt-30b/blob/main/config.json
opt_30b_config = ModelConfig(model='facebook/opt-30B',
    hidden_size=7168, num_attention_heads=56,
    intermediate_size=4*7168, num_decoder_layers=48,
    max_model_len=2*1024, vocab_size=50272, hidden_act="relu",
)

# https://huggingface.co/intlsy/opt-175b-hyperparam/blob/main/config.json
gpt3_config = ModelConfig(model='openai/GPT3-175B',
    hidden_size=12288, num_attention_heads=96, num_ffi = 1,
    intermediate_size=4*12288, num_decoder_layers=96,
    max_model_len=2*1024, vocab_size=50272, hidden_act="relu",
)

######## LLaMA Models ########

# https://huggingface.co/Secbone/llama-33B-instructed/blob/main/config.json
llama_33b_config = ModelConfig(model='meta-llama/Llama-33B',
    hidden_size=6656, num_attention_heads=52, num_ffi = 2,
    intermediate_size=17920, num_decoder_layers=60,
    vocab_size=32000, max_model_len=2*1024, hidden_act="silu",
)

# https://huggingface.co/meta-llama/Llama-2-7b-hf/blob/main/config.json
llama2_7b_config = ModelConfig(model='meta-llama/Llama-2-7B',
    hidden_size=4096, num_attention_heads=32, num_ffi = 2,
    intermediate_size=11008, num_decoder_layers=32,
    vocab_size=32000, max_model_len=4*1024, hidden_act="silu",
)

# https://huggingface.co/meta-llama/Llama-2-13b-hf/blob/main/config.json
llama_13b_config = ModelConfig(model='meta-llama/Llama-2-13B',
    hidden_size=5120, num_attention_heads=40, num_ffi = 2,
    intermediate_size=13824, num_decoder_layers=40,
    vocab_size=32000, max_model_len=4*1024, hidden_act="silu",
)

# https://huggingface.co/meta-llama/Llama-2-70b-hf/blob/main/config.json
llama_70b_config = ModelConfig(model='meta-llama/Llama-2-70B',
    hidden_size=8192, num_attention_heads=64,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=28672, num_decoder_layers=80,
    vocab_size=32000, max_model_len=4*1024, hidden_act="silu",
)

# https://huggingface.co/meta-llama/Meta-Llama-3.1-8B/blob/main/config.json
llama3_8b_config = ModelConfig(model='meta-llama/Llama-3.1-8B',
    hidden_size=4096, num_attention_heads=32,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=14336, num_decoder_layers=32,
    vocab_size=128256, max_model_len=128*1024, hidden_act="silu",

)

# https://huggingface.co/meta-llama/Llama-3.1-70B-Instruct/blob/main/config.json
llama3_70b_config = ModelConfig(model='meta-llama/Llama-3.1-70B',
    hidden_size=8192, num_attention_heads=64,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=28672, num_decoder_layers=80,
    vocab_size=128256, max_model_len=128*1024, hidden_act="silu",
)

# https://huggingface.co/meta-llama/Meta-Llama-3.1-405B
llama3_405b_config = ModelConfig(model='meta-llama/Llama-3.1-405B',
    hidden_size=16384, num_attention_heads=128,
    num_key_value_heads=16, num_ffi = 2,
    intermediate_size=3.25*16384, num_decoder_layers=126,
    vocab_size=128256, max_model_len=128*1024, hidden_act="silu",

)

# https://huggingface.co/meta-llama/Llama-3.2-1B/blob/main/config.json
llama3_2_1b_config = ModelConfig(model='meta-llama/Llama-3.2-1B',
    hidden_size=2048, num_attention_heads=32,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=4*2048, num_decoder_layers=16,
    vocab_size=128256, max_model_len=128*1024, hidden_act="silu",
)

# https://huggingface.co/meta-llama/Llama-3.2-3B/blob/main/config.json
llama3_2_3b_config = ModelConfig(model='meta-llama/Llama-3.2-3B',
    hidden_size=3072, num_attention_heads=24,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=4*2048, num_decoder_layers=28,
    vocab_size=128256, max_model_len=128*1024, hidden_act="silu",
)

meta_models = get_all_model_configs(__name__)
meta_models.update({
    'facebook/opt-1.3b': opt_1b_config,
    'facebook/opt-125m': opt_125m_config,
    'facebook/opt-13b': opt_13b_config,
    'facebook/opt-175b': gpt3_config,
    'facebook/opt-30B': opt_30b_config,
    'facebook/opt-350m': opt_350m_config,
    'facebook/opt-7b': opt_7b_config,
    'gpt-3_1b': opt_1b_config,
    'gpt-2': gpt2_config,
    'gpt-3_7b': opt_7b_config,
    'opt_125m': opt_125m_config,
    'opt_13b': opt_13b_config,
    'opt_175b': gpt3_config,
    'opt_1b': opt_1b_config,
    'opt_30b': opt_30b_config,
    'opt_350m': opt_350m_config,
    'opt_7b': opt_7b_config,
    'llama_33b': llama_33b_config,
    'llama2_7b': llama2_7b_config,
    'llama2_13b': llama_13b_config,
    'llama2_70b': llama_70b_config,
    'llama3_8b': llama3_8b_config,
    'llama_405b': llama3_405b_config,
    'meta-llama/Llama-3.2-1B': llama3_2_1b_config,
    'meta-llama/Llama-3.2-3B': llama3_2_3b_config,
    'meta-llama/llama-2-13b': llama_13b_config,
    'meta-llama/llama-2-33b': llama_33b_config,
    'meta-llama/llama-2-70b': llama_70b_config,
    'meta-llama/llama-2-7b': llama2_7b_config,
    'meta-llama/meta-llama-3.1-405b': llama3_405b_config,
    'meta-llama/meta-llama-3.1-70b': llama_70b_config,
    'meta-llama/meta-llama-3.1-8b': llama3_8b_config,
    'gpt-3': gpt3_config,
    'openai/gpt-3': gpt3_config,
})


from ..default_models import ModelConfig, get_all_model_configs
##### Gemma Models #####
# https://huggingface.co/google/gemma-2b-it/blob/main/config.json
gemma_2b_config = ModelConfig(model='google/gemma-2B',
    hidden_size=2048, num_attention_heads=8, num_ffi = 2,
    num_key_value_heads=1, head_dim=256,
    intermediate_size=16384, num_decoder_layers=18,
    vocab_size=256000, max_model_len=8*1024, hidden_act="gelu",
)

# https://huggingface.co/google/gemma-7b-it/blob/main/config.json
gemma_7b_config = ModelConfig(model='google/gemma-7B',
    hidden_size=3072, num_attention_heads=16, num_ffi = 2,
    intermediate_size=24576, num_decoder_layers=28, head_dim=256,
    vocab_size=256000, max_model_len=8*1024, hidden_act="gelu",
)

# https://huggingface.co/google/gemma-2-2b/blob/main/config.json
gemma2_2b_config = ModelConfig(model='google/gemma-2-2B',
    hidden_size=2304, num_attention_heads=8, num_ffi = 2,
    num_key_value_heads=4, head_dim=256,
    intermediate_size=9216, num_decoder_layers=26,
    vocab_size=256000, max_model_len=8*1024, hidden_act="gelu_pytorch_tanh",
    sliding_window=4096,
)

# https://huggingface.co/google/gemma-2-9b/blob/main/config.json
gemma2_9b_config = ModelConfig(model='google/gemma-2-9B',
    hidden_size=3584, num_attention_heads=16, num_ffi = 2,
    num_key_value_heads=8, head_dim=256,
    intermediate_size=14336, num_decoder_layers=42,
    vocab_size=256000, max_model_len=8*1024, hidden_act="gelu_pytorch_tanh",
    sliding_window=4096,
)

# https://huggingface.co/google/gemma-2-27b-it/blob/main/config.json
gemma2_27b_config = ModelConfig(model='google/gemma-2-27B',
    hidden_size=4608, num_attention_heads=32, num_ffi = 2,
    num_key_value_heads=16, head_dim=128,
    intermediate_size=36864, num_decoder_layers=46,
    vocab_size=256000, max_model_len=8*1024, hidden_act="gelu_pytorch_tanh",
    sliding_window=4096,
)






####################
palm_config = ModelConfig(model='google/palm',
        hidden_size=18432, num_attention_heads=48, num_ffi = 1,
        intermediate_size=4*18432, num_decoder_layers=118
)

google_models = get_all_model_configs(__name__)

google_models.update({
    'gemma2_27b': gemma2_27b_config,
    'gemma2_9b': gemma2_9b_config,
    'gemma2_2b': gemma2_2b_config,
    'gemma_2b': gemma_2b_config,
    'gemma_7b': gemma_7b_config,
    'google/gemma-2-27b-it': gemma2_27b_config,
    'google/gemma-2-9b': gemma2_9b_config,
    'google/gemma-2-2b': gemma2_2b_config,
    'google/gemma-2b-it': gemma_2b_config,
    'google/gemma-7b-it': gemma_7b_config,
    'google/palm': palm_config,
    'palm': palm_config,
})


from ..default_models import ModelConfig, get_all_model_configs

##### Qwen1 Models ########
# https://huggingface.co/Qwen/Qwen-1_8B/blob/main/config.json
qwen1_8b_config = ModelConfig(model='Qwen/Qwen-1_8B',
    hidden_size=2048, num_attention_heads=16,
    num_key_value_heads=16, num_ffi = 1,
    intermediate_size=11008, num_decoder_layers=24,
    vocab_size=151936, max_model_len=8*1024, sliding_window=32*1024,hidden_act="silu")

# https://huggingface.co/Qwen/Qwen-7B/blob/main/config.json
qwen1_7b_config = ModelConfig(model='Qwen/Qwen-7B',
    hidden_size=4096, num_attention_heads=32,
    num_key_value_heads=32, num_ffi = 1,
    intermediate_size=22016, num_decoder_layers=32,
    vocab_size=151936, max_model_len=8*1024, sliding_window=32*1024,hidden_act="silu")

# https://huggingface.co/Qwen/Qwen-14B/blob/main/config.json
qwen1_14b_config = ModelConfig(model='Qwen/Qwen-14B',
    hidden_size=5120, num_attention_heads=40,
    num_key_value_heads=40, num_ffi = 1,
    intermediate_size=27392, num_decoder_layers=40,
    vocab_size=152064, max_model_len=2*1024, sliding_window=32*1024,hidden_act="silu")

# https://huggingface.co/Qwen/Qwen-72B/blob/main/config.json
qwen1_72b_config = ModelConfig(model='Qwen/Qwen-72B',
    hidden_size=8192, num_attention_heads=64,
    num_key_value_heads=64, num_ffi = 1,
    intermediate_size=49152, num_decoder_layers=80,
    vocab_size=152064, max_model_len=32*1024, sliding_window=32*1024,hidden_act="silu")

##### Qwen1.5 Models ########
# https://huggingface.co/Qwen/Qwen1.5-0.5B/blob/main/config.json
qwen1_5_0_5b_config = ModelConfig(model='Qwen/Qwen1.5-0.5B',
    hidden_size=1024, num_attention_heads=16,
    num_key_value_heads=16, num_ffi = 2,
    intermediate_size=2816, num_decoder_layers=24,
    vocab_size=151936, max_model_len=32*1024, sliding_window=32*1024,hidden_act="silu")

# https://huggingface.co/Qwen/Qwen1.5-1.8B/blob/main/config.json
qwen1_5_1_8b_config = ModelConfig(model='Qwen/Qwen1.5-1.8B',
    hidden_size=2048, num_attention_heads=16,
    num_key_value_heads=16, num_ffi = 2,
    intermediate_size=5504, num_decoder_layers=24,
    vocab_size=151936, max_model_len=32*1024, sliding_window=32*1024,hidden_act="silu")

# https://huggingface.co/Qwen/Qwen1.5-4B/blob/main/config.json
qwen1_5_4b_config = ModelConfig(model='Qwen/Qwen1.5-4B',
    hidden_size=2560, num_attention_heads=20,
    num_key_value_heads=20, num_ffi = 2,
    intermediate_size=6912, num_decoder_layers=40,
    vocab_size=151936, max_model_len=32*1024, sliding_window=32*1024,hidden_act="silu")

# https://huggingface.co/Qwen/Qwen1.5-7B/blob/main/config.json
qwen1_5_7b_config = ModelConfig(model='Qwen/Qwen1.5-7B',
    hidden_size=4096, num_attention_heads=32,
    num_key_value_heads=32, num_ffi = 2,
    intermediate_size=11008, num_decoder_layers=32,
    vocab_size=151936, max_model_len=32*1024, sliding_window=32*1024,hidden_act="silu")

# https://huggingface.co/Qwen/Qwen1.5-14B/blob/main/config.json
qwen1_5_14b_config = ModelConfig(model='Qwen/Qwen1.5-14B',
    hidden_size=5120, num_attention_heads=40,
    num_key_value_heads=40, num_ffi = 2,
    intermediate_size=13696, num_decoder_layers=40,
    vocab_size=152064, max_model_len=32*1024, sliding_window=32*1024,hidden_act="silu")

# https://huggingface.co/Qwen/Qwen1.5-32B/blob/main/config.json
qwen1_5_32b_config = ModelConfig(model='Qwen/Qwen1.5-32B',
    hidden_size=5120, num_attention_heads=64,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=27392, num_decoder_layers=64,
    vocab_size=152064, max_model_len=32*1024, sliding_window=32*1024,hidden_act="silu")

# https://huggingface.co/Qwen/Qwen1.5-72B/blob/main/config.json
qwen1_5_72b_config = ModelConfig(model='Qwen/Qwen1.5-72B',
    hidden_size=8192, num_attention_heads=64,
    num_key_value_heads=64, num_ffi = 2,
    intermediate_size=24576, num_decoder_layers=80,
    vocab_size=152064, max_model_len=32*1024, sliding_window=32*1024,hidden_act="silu")

# https://huggingface.co/Qwen/Qwen1.5-110B/blob/main/config.json
qwen1_5_110b_config = ModelConfig(model='Qwen/Qwen1.5-110B',
    hidden_size=8192, num_attention_heads=64,
    num_key_value_heads=64, num_ffi = 2,
    intermediate_size=49152, num_decoder_layers=80,
    vocab_size=152064, max_model_len=32*1024, sliding_window=32*1024,hidden_act="silu")

# https://huggingface.co/Qwen/Qwen1.5-MoE-A2.7B/blob/main/config.json
qwen1_5_moe_2_7b_config = ModelConfig(model='Qwen/Qwen1.5-MoE-A2.7B',
    hidden_size=2048, num_attention_heads=16,
    num_key_value_heads=16, num_ffi = 2,
    intermediate_size=5632, num_decoder_layers=40,
    expert_top_k=4, num_experts=60, moe_layer_freq=1,
    moe_intermediate_size=1408, shared_expert_intermediate_size=5632,
    vocab_size=151936, max_model_len=8*1024, sliding_window=32*1024, hidden_act="silu",
)
## TODO: account for shared expert, shared account is regular MLP which is always added.
## MLP in this case is shared_expert_intermediate_size + Activated*moe_intermediate_size

##### Qwen2 Models ########
# https://huggingface.co/Qwen/Qwen2-0.5B/blob/main/config.json
qwen2_0_5b_config = ModelConfig(model='Qwen/Qwen2-0.5B',
    hidden_size=896, num_attention_heads=14,
    num_key_value_heads=2, num_ffi = 2,
    intermediate_size=4864, num_decoder_layers=24,
    vocab_size=151936, max_model_len=128*1024, sliding_window=128*1024,hidden_act="silu")

# https://huggingface.co/Qwen/Qwen2.5-1.5B-Instruct/blob/main/config.json
qwen2_1_5b_config = ModelConfig(model='Qwen/Qwen2.5-1.5B',
    hidden_size=1536, num_attention_heads=12,
    num_key_value_heads=2, num_ffi = 2,
    intermediate_size=8960, num_decoder_layers=28,
    vocab_size=151936, max_model_len=32*1024, sliding_window=32*1024,hidden_act="silu")

# https://huggingface.co/Qwen/Qwen2-3B-Instruct/blob/main/config.json
qwen2_3b_config = ModelConfig(model='Qwen/Qwen2-3B',
    hidden_size=2048, num_attention_heads=16,
    num_key_value_heads=2, num_ffi = 2,
    intermediate_size=11008, num_decoder_layers=36,
    vocab_size=151936, max_model_len=32*1024, sliding_window=32*1024,hidden_act="silu")

# https://huggingface.co/Qwen/Qwen2-7B-Instruct/blob/main/config.json
qwen2_7b_config = ModelConfig(model='Qwen/Qwen2-7B',
    hidden_size=3584, num_attention_heads=28,
    num_key_value_heads=4, num_ffi = 2,
    intermediate_size=18944, num_decoder_layers=28,
    vocab_size=152064, max_model_len=32*1024, sliding_window=32*1024,hidden_act="silu")

# https://huggingface.co/Qwen/Qwen2-14B-Instruct/blob/main/config.json
qwen2_14b_config = ModelConfig(model='Qwen/Qwen2-14B',
    hidden_size=5120, num_attention_heads=40,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=13824, num_decoder_layers=48,
    vocab_size=152064, max_model_len=32*1024, sliding_window=32*1024,hidden_act="silu")

# https://huggingface.co/Qwen/Qwen2-32B-Instruct/blob/main/config.json
qwen2_32b_config = ModelConfig(model='Qwen/Qwen2-32B',
    hidden_size=5120, num_attention_heads=40,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=27648, num_decoder_layers=64,
    vocab_size=152064, max_model_len=32*1024, sliding_window=32*1024,hidden_act="silu")

# https://huggingface.co/Qwen/Qwen2-72B-Instruct/blob/main/config.json
qwen2_72b_config = ModelConfig(model='Qwen/Qwen2-72B',
    hidden_size=8192, num_attention_heads=64,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=29568, num_decoder_layers=80,
    vocab_size=152064, max_model_len=32*1024, sliding_window=32*1024,hidden_act="silu")


alibaba_models = get_all_model_configs(__name__)




import numpy as np

class ParallelismConfig():
    r"""
    This is the configuration class to store the configuration of a Model Splitting.
    It is used to instantiate an LLM into multiple parallel units
    according to the specified arguments, defining the degree of various parallelism.
    Args:

    """
    def __init__(
        self,
        tensor_parallel=1,
        pipeline_parallel=1,
        data_parallel=1,
        expert_parallel=1,
        sequence_parallel=1,
        **kwargs,
    ):
        self.tensor_parallel = tensor_parallel
        self.pipeline_parallel = pipeline_parallel
        self.data_parallel = data_parallel
        self.expert_parallel = expert_parallel
        self.sequence_parallel = sequence_parallel
        self.total_chips = np.prod([
                            self.data_parallel,
                            self.expert_parallel,
                            self.sequence_parallel,
                            self.pipeline_parallel,
                            self.tensor_parallel])

        super().__init__(**kwargs)

    def __str__(self):
        return str(vars(self))


from GenZ.Models import ModelConfig, ResidencyInfo, OpType, parse_einsum_expression
from GenZ.parallelism import ParallelismConfig

def mamda_ssn_slow(model_config:ModelConfig, parallelism_config:ParallelismConfig, input_sequence_length:int):
    tp = parallelism_config.tensor_parallel * parallelism_config.expert_parallel
    sp = parallelism_config.sequence_parallel
    dp = parallelism_config.data_parallel

    L = input_sequence_length ## input Seq Len
    ## Mamba parameters
    D = model_config.hidden_size
    S = model_config.mamba_d_state
    C = model_config.mamba_d_conv
    F = D * model_config.mamba_expand
    R = model_config.mamba_dt_rank

    """
    u: r(B F L)
    delta: r(B F L)
    A: r(F S)
    B: r(B S L)
    C: r(B S L)
    D: r(F)
    z: r(B F L)
    delta_bias: r(F), fp32

    out: r(B F L)

    deltaA = torch.exp(torch.einsum('bdl,dn->bdln', delta, A))
    deltaB_u = torch.einsum('bdl,bnl,bdl->bdln', delta, B, u)
    for i in range(L):
        x = deltaA[:, :, i] * x + deltaB_u[:, :, i]
        y = torch.einsum('bdn,bn->bd', x, C[:, :, i])
        ys.append(y)
    y = torch.stack(ys, dim=2) # (batch dim L)
    out = y + u * rearrange(D, "d -> d 1")
    out = out * F.silu(z)
    """

    Layers = []
    ## A and intermediate tensors on chip
    deltaA = [["deltaA",'bdl,dn->bdln',parse_einsum_expression('bdl,dn->bdln', ('b',F,L), (F,S), ('b',F,L,S)), 1, 1, 1, ResidencyInfo.BC_onchip, OpType.EINSUM]]
    ## output off chip
    deltaB = [["deltaB",'bdl,bls->bdls',parse_einsum_expression('bdl,bls->bdls', ('b',F,L), ('b',L,S), ('b',F,L,S)), 1, 1, 1, ResidencyInfo.C_onchip, OpType.EINSUM]]
    ## U and output on-chip
    deltaB_u = [["deltaBu",'bdls,bdl->bdls',parse_einsum_expression('bdls,bdl->bdls', ('b',F,L,S), ('b',F,L), ('b',F,L,S)), 1, 1, 1, ResidencyInfo.All_offchip, OpType.EINSUM]]
    Layers += deltaA + deltaB + deltaB_u
    # for _ in range(L):
    Layers += [["x calc",'lbfs,lbfs->lbfs',parse_einsum_expression('lbfs,lbfs->lbfs', (L,'b',F,S), (L,'b',F,S), (L,'b',F,S)), 1, 1, 1, ResidencyInfo.All_onchip, OpType.EINSUM]]
    Layers += [["y calc",'lbfs,lbs->lbf',parse_einsum_expression('lbfs,lbs->lbf', (L,'b',F,S), (L,'b',S), (L,'b',F)), 1, 1, 1, ResidencyInfo.All_onchip, OpType.EINSUM]]
    Layers += [["D addition",'blf,blf->blf',parse_einsum_expression('blf,blf->blf', ('b',L,F), ('b',L,F), ('b',L,F)), 1, 1, 1, ResidencyInfo.All_onchip, OpType.EINSUM]] 
    Layers += [["out mult z",'blf,blf->blf',parse_einsum_expression('blf,blf->blf', ('b',L,F), ('b',L,F), ('b',L,F)), 1, 1, 1, ResidencyInfo.All_onchip, OpType.EINSUM]] 

    return Layers

def mamba_prefill(model_config:ModelConfig, parallelism_config:ParallelismConfig, input_sequence_length:int):
    tp = parallelism_config.tensor_parallel * parallelism_config.expert_parallel
    sp = parallelism_config.sequence_parallel
    dp = parallelism_config.data_parallel

    L  = input_sequence_length ## input Seq Len

    D = model_config.hidden_size
    Df = model_config.intermediate_size

    ## Mamba parameters
    S = model_config.mamba_d_state
    C = model_config.mamba_d_conv
    F = D * model_config.mamba_expand
    R = model_config.mamba_dt_rank

    # assert H % tensor_parallel == 0, f'Heads should be equally divisible, H:{H}, TP:{tensor_parallel}'
    Df = max(Df//tp,1)

    in_proj =      [["Inproj",2*F, L, D, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]        ## BLD * D2F = BL2F
    conv_1d =      [["Conv",F, F, L, C, 1, ResidencyInfo.All_offchip, OpType.CONV1D]]         ## BLF conv FC -> BLF
    dbc_proj =     [["BC proj",R+2*S, L, F, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]
    xt_proj =      [["xt proj",F, L, R, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]
    ssn =   mamda_ssn_slow(model_config, parallelism_config, input_sequence_length)

    output =       [["Out proj",D, L, F, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]

    return in_proj + conv_1d + dbc_proj + xt_proj + ssn + output


def mamba_decode(model_config:ModelConfig, parallelism_config:ParallelismConfig,
                input_sequence_length:int):

    L  = input_sequence_length ## input Seq Len

    tp = parallelism_config.tensor_parallel * parallelism_config.expert_parallel
    sp = parallelism_config.sequence_parallel
    dp = parallelism_config.data_parallel

    ## Mamba parameters
    D = model_config.hidden_size
    S = model_config.mamba_d_state
    C = model_config.mamba_d_conv
    F = D * model_config.mamba_expand
    R = model_config.mamba_dt_rank

    in_proj =      [["Inproj",2*F, 1, D, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]        ## BLD * D2F = BL2F
    conv_1d =      [["Conv", F, F, 1, C, 1, ResidencyInfo.All_offchip, OpType.CONV1D]]         ## BLF conv FC -> BLF
    dbc_proj =     [["BC proj",R+2*S, 1, F, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]
    xt_proj =      [["xt proj",F, 1, R, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]
    ssn =   mamda_ssn_slow(model_config, parallelism_config, 1)
    output =       [["Out proj",D, 1, F, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]

    return in_proj + conv_1d + dbc_proj + xt_proj + ssn + output


import pytest
from GenZ import prefill_moddeling, get_model_df, get_configs, System, create_inference_moe_prefill_layer, get_AR_time
import os
import pandas as pd
import numpy as np

def test_dense_LLM_prefill():
    # Generate the current result
    TPU = System(flops=300, offchip_mem_bw=1200, compute_efficiency=0.8, memory_efficiency=0.8, bits='bf16')
    Model = 'gpt-2'
    prefill_output = prefill_moddeling(model = Model, batch_size = 1, input_tokens = 4096,
                                system_name = TPU, bits='bf16', tensor_parallel = 1, pipeline_parallel = 1, debug=False)

    ref_latency = 5.476083302399999
    ref_throughput = 182.6122695324468
    ref_runtime_breakdown = [2.8991029248, 2.5769803776, 0.0]
    assert np.allclose([prefill_output['Latency'], prefill_output['Throughput'], prefill_output['Runtime_breakdown'][0], prefill_output['Runtime_breakdown'][1], prefill_output['Runtime_breakdown'][2]],
                        [ref_latency, ref_throughput, ref_runtime_breakdown[0], ref_runtime_breakdown[1], ref_runtime_breakdown[2]])

def test_dense_LLM_prefill_with_tensor_parallel():
    # Generate the current result
    TPU = System(flops=300, offchip_mem_bw=1200, compute_efficiency=0.8, memory_efficiency=0.8, bits='bf16',
                interchip_link_bw=50, interchip_link_latency=1)
    Model = 'gpt-2'
    prefill_output = prefill_moddeling(model = Model, batch_size = 1, input_tokens = 4096,
                                system_name = TPU, bits='bf16', tensor_parallel = 4, pipeline_parallel = 1, debug=False)

    ref_latency = 5.886358809914
    ref_throughput = 169.884309178665
    ref_runtime_breakdown = [0.759363715514, 0.6442450944, 4.48275]
    assert np.allclose([prefill_output['Latency'], prefill_output['Throughput'], prefill_output['Runtime_breakdown'][0], prefill_output['Runtime_breakdown'][1], prefill_output['Runtime_breakdown'][2]],
                        [ref_latency, ref_throughput, ref_runtime_breakdown[0], ref_runtime_breakdown[1], ref_runtime_breakdown[2]])


import numpy as np
from GenZ.LLM_inference import decode_moddeling, prefill_moddeling, get_best_parallization_strategy, get_inference_system

def get_minimum_system_size(
        stage='decode', model='llama2_7b', max_batch_size = 1, beam_size = 1,
        input_tokens = 2000, output_tokens = 256,
        system_name = {'Flops': 200, 'Memory_size': 32, 'Memory_BW': 1000, 'ICN': 300 , 'real_values':True},
        bits='bf16', debug=False
        ):

    if stage=='prefill':
        model_df, summary_table = prefill_moddeling(model = model, batch_size = max_batch_size,
                            input_tokens = input_tokens, model_profilling=True,
                            tensor_parallel = 1, pipeline_parallel = 1, bits=bits, debug=debug)
    elif stage=='decode':
        model_df, summary_table = decode_moddeling(model = model, batch_size = max_batch_size, Bb = beam_size ,
                            input_tokens = input_tokens, output_tokens = output_tokens, model_profilling=True,
                            tensor_parallel = 1, pipeline_parallel = 1, bits=bits, debug=debug)
    Total_memory_required = (summary_table.loc[0,'Total Weights (MB)'] + summary_table.loc[0,'KV Cache (MB)'])   ## MBs
    system = get_inference_system(system_name =system_name , bits = bits)
    Node_memory_size = system.get_off_chip_mem_size()       ## In MBs
    Num_nodes = Total_memory_required / Node_memory_size
    if Num_nodes <= 1:
        Num_nodes = 1
    else:
        Num_nodes = int(np.power(2,np.ceil(np.log2(Num_nodes))))
    ##### RE verifying the memory sizes
    while 1:
        try:    ## Check if model fits
            _ = get_best_parallization_strategy(stage=stage, model=model, total_nodes=Num_nodes, batch_size = max_batch_size, beam_size = beam_size,
                        input_tokens = input_tokens, output_tokens = output_tokens,
                        system_name = system_name, bits=bits, debug=debug).sort_values(by='Tokens/s', ascending=False)
            break
        except: ## If model doesn't fit, have 2x number of cores.
            Num_nodes *= 2

    return Num_nodes


from .utils import (OpType, ResidencyInfo, CollectiveType, parse_einsum_expression)
from .default_models import ModelConfig, MODEL_DICT
from .get_language_model import (
    get_configs,
    create_inference_moe_prefill_layer,
    create_inference_moe_decode_layer,
    create_inference_mamba_prefix_model,
    create_inference_mamba_decode_model,
    create_full_prefill_model,
    create_full_decode_model,
)


from typing import Any, Dict

system_configs: Dict[str, Dict[str, Any]] = {
    'A100_40GB_GPU' : {'Flops': 312, 'Memory_size': 40, 'Memory_BW': 1600, 'ICN': 150 , 'real_values':True},
    'A100_80GB_GPU' : {'Flops': 312, 'Memory_size': 80, 'Memory_BW': 2039, 'ICN': 150 , 'real_values':True},
    'H100_GPU'  : {'Flops': 989, 'Memory_size': 80, 'Memory_BW': 3400, 'ICN': 450 , 'real_values':True},
    'GH200_GPU' : {'Flops': 1979, 'Memory_size': 144, 'Memory_BW': 4900, 'ICN': 450 , 'real_values':True},
    'TPUv5e' :  {'Flops': 197, 'Memory_size': 16, 'Memory_BW': 820, 'ICN': 50 , 'real_values':True},
    'TPUv4' : {'Flops': 275, 'Memory_size': 32, 'Memory_BW': 1200, 'ICN': 50 , 'real_values':True},
    'MI300X' : {'Flops': 1307, 'Memory_size': 192, 'Memory_BW': 5300, 'ICN': 400 , 'real_values':True},
    # https://www.intel.com/content/www/us/en/content-details/817486/intel-gaudi-3-ai-accelerator-white-paper.html
    'Gaudi3' : {'Flops': 1835, 'Memory_size': 128, 'Memory_BW': 3675, 'ICN': 300 , 'real_values':True},
}


from GenZ.unit import Unit
import GenZ.operators as operators
from GenZ.operator_base import op_type_dicts
from GenZ.system import System
import pandas as pd
import numpy as np
import os
from GenZ.Models import OpType, ResidencyInfo

from GenZ.LLM_inference.utils import RuntimeBreakdown

def get_attn_index(df:pd.DataFrame):
    ret = []
    for idx in range(len(df)):
        if 'Attend' in df.loc[idx, 'Op Type'] or 'Logit' in df.loc[idx, 'Op Type']:
            ret.append(idx)
    return ret

def get_summary_table(df:pd.DataFrame, unit = Unit(), model_characterstics:bool=False):

    attn_idx = get_attn_index(df)

    total_macs = 0
    total_data = 0
    kv_cache = 0
    total_weights = 0
    unused_weights = 0
    total_latencies = 0
    total_cycles = 0
    total_attn_latencies = 0
    total_linear_latencies = 0
    total_comm_latencies = 0

    multiplier = 1
    for i in range(len(df)):
        if df.loc[i,'Op Type'] == 'Repeat':
            multiplier *= df.loc[i,'Dimension']
        elif df.loc[i,'Op Type'] == 'EndRepeat':
            multiplier /= df.loc[i,'Dimension']
        else:
            total_macs += df.loc[i,f'Num ops ({unit.unit_flop})'] * multiplier
            total_data += (df.loc[i,f'Input_a ({unit.unit_mem})'] + df.loc[i,f'Input_w ({unit.unit_mem})'] + df.loc[i,f'Output ({unit.unit_mem})']) * multiplier
            if i in attn_idx:
                kv_cache += df.loc[i,f'Input_w ({unit.unit_mem})'] * multiplier
            else:
                total_weights += df.loc[i,f'Input_w ({unit.unit_mem})'] * multiplier
                if df.loc[i, f'Num ops ({unit.unit_flop})'] == 0:
                    unused_weights += df.loc[i,f'Input_w ({unit.unit_mem})'] * multiplier

            if model_characterstics == False:
                total_latencies += df.loc[i,f'Latency ({unit.unit_time})'] * multiplier
                total_cycles += df.loc[i,'Cycles'] * multiplier
                if i in attn_idx:
                    total_attn_latencies += df.loc[i,f'Latency ({unit.unit_time})'] * multiplier
                elif 'GEMM' in df.loc[i, 'Op Type']:
                    total_linear_latencies += df.loc[i,f'Latency ({unit.unit_time})'] * multiplier
                elif 'Sync' in df.loc[i, 'Op Type']:
                    total_comm_latencies += df.loc[i,f'Latency ({unit.unit_time})'] * multiplier

    max_memory_footprint = max([df.loc[i, f'Input_a ({unit.unit_mem})'] + df.loc[i, f'Input_w ({unit.unit_mem})'] + df.loc[i, f'Output ({unit.unit_mem})'] for i in range(len(df))])


    ret = {
            f'MACs ({unit.unit_flop})': [total_macs],
            f'Total Data ({unit.unit_mem})': [total_data],
            f'Total Weights ({unit.unit_mem})': [total_weights],
            f'Unused Weights ({unit.unit_mem})': [unused_weights],
            f'KV Cache ({unit.unit_mem})': [kv_cache],
            f'On-chip Memory Footprint ({unit.unit_mem})': [max_memory_footprint],
        }
    if model_characterstics == False:
        ret.update({
            f'Latency ({unit.unit_time})': [total_latencies],
            'Cycles': [total_cycles],
            f'Attn Latency ({unit.unit_time})': [total_attn_latencies],
            f'Linear Latency ({unit.unit_time})': [total_linear_latencies],
            f'Comm Latency ({unit.unit_time})': [total_comm_latencies]
        })


    return pd.DataFrame.from_dict(ret)

def simplify_df(df:pd.DataFrame):
    unit = Unit()
    column_to_update = [f'Latency ({unit.unit_time})',
                f'Compute time ({unit.unit_time})', f'Memory time ({unit.unit_time})', f'Communication time ({unit.unit_time})',
                f'Cycles', f'Compute cycle', f'Memory cycle', f'Communication cycle',
                f'Num ops ({unit.unit_flop})', f'Input_a ({unit.unit_mem})', f'Input_w ({unit.unit_mem})', f'Output ({unit.unit_mem})', f'Total Data ({unit.unit_mem})',
                ]
    column_no_change = df.columns.difference(column_to_update).tolist()

    multiplier = 1
    new_df = pd.DataFrame(columns=df.columns)
    for i in range(len(df)):
        if df.loc[i,'Op Type'] == 'Repeat':
            multiplier *= df.loc[i,'Dimension']
        elif df.loc[i,'Op Type'] == 'EndRepeat':
            multiplier /= df.loc[i,'Dimension']
        else:
            new_row = df.loc[i, column_no_change].copy()
            for col in column_to_update:
                new_row[col] = df.loc[i, col] * multiplier
            new_df = pd.concat([new_df, pd.DataFrame([new_row])], ignore_index=True)
    return new_df

def get_runtime_breakdown(df:pd.DataFrame) -> RuntimeBreakdown:
    df = simplify_df(df)
    unit = Unit()
    runtime_breakdown = RuntimeBreakdown()
    assert 'Layer Name' in df.columns, "Layer Name not found in the dataframe"
    assert f'Latency ({unit.unit_time})' in df.columns, "Latency (ms) not found in the dataframe"

    possible_layer_names = ['embeddings', 'classifier', 'Emb_AR', 'classifier_AG', 
                            'QKV', 'Out Proj',
                            'Logit', 'Attend',
                            'Logit Pre', 'Logit Suf',
                            'Attend Pre', 'Attend Suf',
                            'Gate', 'up+gate', 'down',
                            'Message Pass', 'MHA AR', 'Gate AR',
                            'Dispatch A2A', 'Collect A2A', 'FFN AR']

    for i in range(len(df)):
        layer_name = df.loc[i, 'Layer Name']
        layer_latency = df.loc[i, f'Latency ({unit.unit_time})']
        if layer_name in ['embeddings', 'classifier']:
            runtime_breakdown.Embedding += layer_latency
        elif layer_name in ['QKV', 'Out Proj']:
            runtime_breakdown.MHA += layer_latency
            runtime_breakdown.QKVO_layers += layer_latency
        elif layer_name in ['Logit', 'Attend', 'Logit Pre', 'Logit Suf',
                            'Attend Pre', 'Attend Suf',]:
            runtime_breakdown.MHA += layer_latency
            runtime_breakdown.LA_layers += layer_latency
        elif layer_name in ['Gate', 'up+gate', 'down']:
            runtime_breakdown.FFN += layer_latency
        elif layer_name in ['Message Pass']:
            runtime_breakdown.Collective += layer_latency
            runtime_breakdown.Send_Recv_time += layer_latency
        elif layer_name in ['MHA AR']:
            runtime_breakdown.MHA += layer_latency
            runtime_breakdown.Collective += layer_latency
            runtime_breakdown.AR_time += layer_latency
        elif layer_name in ['Gate AR', 'FFN AR']:
            runtime_breakdown.FFN += layer_latency
            runtime_breakdown.Collective += layer_latency
            runtime_breakdown.AR_time += layer_latency
        elif layer_name in ['Dispatch A2A', 'Collect A2A']:
            runtime_breakdown.Collective += layer_latency
            runtime_breakdown.A2A_time += layer_latency
            runtime_breakdown.FFN += layer_latency
        elif layer_name in ['Emb_AR', 'classifier_AG']:
            runtime_breakdown.AR_time += layer_latency
            runtime_breakdown.Embedding += layer_latency
            runtime_breakdown.Collective += layer_latency
        else:
            raise ValueError(f'Layer Name:{layer_name} not found in the breakdown function')

    return runtime_breakdown

def analysis_model(model_dims, system=None, unit=Unit(), densities = None,intermediate_on_chip=False,
                    beam_size=1, beam_merge=False, model_characterstics=False):
    roofline_list = []
    if densities is None:
        densities = np.ones((len(model_dims), 3), dtype=float)
    for i, (dim, density) in enumerate(zip(model_dims, densities)):

        op_type = op_type_dicts[dim[-1]]
        operators_residency = dim[-2]
        operator = getattr(operators, op_type)
        if beam_merge and (dim[-1] == OpType.Logit_BM_PREFILL or dim[-1] == OpType.Attend_BM_PREFILL):
            dim[1] /= beam_size         ## Batch size is divided by beam size
        operator_instance = operator(dim=dim, density=density)
        # print(density[0],density[1],density[2])
        if (intermediate_on_chip):
            if(op_type == 'Logit'):
                operator_instance.set_mem_pin(output='on')
            elif(op_type == 'Attend'):
                operator_instance.set_mem_pin(input_a='on')

        if operators_residency == ResidencyInfo.A_onchip:
            operator_instance.set_mem_pin(input_a='on')
        elif operators_residency == ResidencyInfo.B_onchip:
            operator_instance.set_mem_pin(input_b='on')
        elif operators_residency == ResidencyInfo.C_onchip:
            operator_instance.set_mem_pin(output='on')
        elif operators_residency == ResidencyInfo.AB_onchip:
            operator_instance.set_mem_pin(input_a='on')
            operator_instance.set_mem_pin(input_b='on')
        elif operators_residency == ResidencyInfo.AC_onchip:
            operator_instance.set_mem_pin(input_a='on')
            operator_instance.set_mem_pin(output='on')
        elif operators_residency == ResidencyInfo.BC_onchip:
            operator_instance.set_mem_pin(input_b='on')
            operator_instance.set_mem_pin(output='on')
        elif operators_residency == ResidencyInfo.All_onchip:
            operator_instance.set_mem_pin(input_a='on')
            operator_instance.set_mem_pin(input_b='on')
            operator_instance.set_mem_pin(output='on')

        if model_characterstics:
            roofline = operator_instance.get_model_characterstics(system=system, unit=unit)
        else:
            roofline = operator_instance.get_roofline(system=system, unit=unit)

        if i==0:
            column = roofline.keys()
        roofline_list.append([roofline[c] for c in column])

    df = pd.DataFrame(np.array(roofline_list,dtype=object), columns=column, dtype=object)

    return df


def get_model_df(model, system=System(), unit=Unit(), batch_size=1, data_path="/tmp/genz/data", intermediate_on_chip=False,
                    beam_size=1, beam_merge=False, model_characterstics=False):
    m_file_path = os.path.join(data_path,"model")
    sparsity_file_path = os.path.join(data_path,"sparsity")
    m_file = os.path.join(m_file_path, model)
    density_file = os.path.join(sparsity_file_path, model)
    df = pd.read_csv(m_file)
    model_defs = df.to_numpy()
    model_defs = np.insert(model_defs, 1, batch_size, axis=1)
    # model_defs = np.append(batch_sizes, model_defs, axis=1)
    def verify_repeat_pairs(model_defs):
        pairs = []
        stack = []
        for idx, row in enumerate(model_defs):
            Repeat_id = row[2]
            if row[-1] == OpType.REPEAT:
                stack.append((idx, Repeat_id))
            elif row[-1] == OpType.ENDREPEAT:
                if stack and stack[-1][1] == Repeat_id:
                    start_idx, _ = stack.pop()
                    pairs.append((start_idx, idx))
                else:
                    raise ValueError(f"Unmatched Endrepeat found or ID mismatch:{Repeat_id}")

        if stack:
            raise ValueError(f"Unmatched Repeat found: {stack[-1]}")

        return pairs
    pairs = verify_repeat_pairs(model_defs)

    new_model_defs = []
    for layer in model_defs:
        # if layer[-1] == OpType.EINSUM:
            new_layer = [int(x) if isinstance(x, (int, float)) else x for x in layer]
            new_model_defs.append(new_layer)
        # else:
        #     new_model_defs.append(layer.astype(int))

    densities = np.ones((len(model_defs), 3), dtype=float)

    return analysis_model(new_model_defs, system, unit, densities, intermediate_on_chip, beam_size, beam_merge, model_characterstics)




import numpy as np

# based on the first letters of the units trying to get values in readable form.
class Unit(object):
    unit_dicts = {'K': 1e3, 'M': 1e6, 'G':1e9, 'T':1e12, 'm': 1e-3, 'u': 1e-6, 'n': 1e-9, 'p': 1e-12}
    binary_dicts = {'K': 2**10, 'M': 2**20, 'G':2**30, 'T':2**40, 'm': 2**-10, 'u': 2**-20, 'n': 2**-30, 'p': 2**-40}

    def __init__(self, unit_mem='MB', unit_compute='Tflops', unit_time='msec', unit_bw='GBsec', unit_freq='MHz',
                unit_energy='pJ', unit_flop='MFLOP'):
        self.unit_mem = unit_mem
        self.unit_compute = unit_compute
        self.unit_time = unit_time
        self.unit_bw = unit_bw
        self.unit_freq = unit_freq
        self.unit_energy = unit_energy
        self.unit_flop = unit_flop

    def get_unit_value(self, type):
        if type =='C':          ## Compute
            unit_value = self.unit_dicts[self.unit_compute[0]]
        elif type == 'M':       ## Memory
            unit_value = self.binary_dicts[self.unit_mem[0]]
        elif type == 'T':       ## Time
            unit_value = self.unit_dicts[self.unit_time[0]]
        elif type == 'BW':      ## Bandwidth
            unit_value = self.binary_dicts[self.unit_bw[0]]
        elif type == 'F':       ## Frequency
            unit_value = self.unit_dicts[self.unit_freq[0]]
        elif type == 'E':       ## Energy
            unit_value = self.unit_dicts[self.unit_energy[0]]
        elif type == 'O':       ## Floting point operations
            unit_value = self.unit_dicts[self.unit_flop[0]]
        else:
            raise ValueError(f'Wrong unit type: {type}')
        return unit_value

    def raw_to_unit(self, data, type='C'):
        unit_value = self.get_unit_value(type=type)
        return data / unit_value

    def unit_to_raw(self, data, type='C'):
        unit_value = self.get_unit_value(type=type)
        return data * unit_value




Model_list = ['llama2_7b', 'mixtral_8x7b', 'LLaMA_70b', 'gpt-3', 'gpt-4']
Model_names_in_plot = ['LLaMA2-07B', 'Mixtral-8x7B', 'LLaMA3-70B', 'GPT3-175B', 'GPT4-1.8T'] ## Only to be used for plot

## Format: Batch, Beam, Input Tokens, Output Tokens, Latency(s)
Prefill_requirements = {
        'QA':(1,4,1000,200,0.2),
        'Chat Bots':(1,2,3000,1000,0.2),
        'QA + RAG':(1,4,10000,200,0.4),
        'Summarization':(1,4,15000,1000,2),
        'Code Generation':(1,4,20000,50,5)
        }

Decode_requirements = {
        'QA':(1,4,1000,200,0.010),
        'Chat Bots':(1,2,3000,1000,0.010),
        'QA + RAG':(1,4,10000,200,0.010),
        'Summarization':(1,4,15000,1000,0.020),
        'Code Generation':(1,4,20000,50,0.020)
        }

system_list = [
'TPUv5e', 'H100', 'MI300X', 'gaudi3',   ## Real Systems
{'Flops':300, 'Memory_BW':600, 'Memory_size':16, 'ICN':400, 'real_values':True},    # QA
{'Flops':900, 'Memory_BW':3600, 'Memory_size':16, 'ICN':400, 'real_values':True},    # Chat Bots
{'Flops':1800, 'Memory_BW':2500, 'Memory_size':24, 'ICN':400, 'real_values':True},    # QA + RAG
{'Flops':750, 'Memory_BW':1875, 'Memory_size':48, 'ICN':400, 'real_values':True},    # Summarization
{'Flops':1000, 'Memory_BW':1250, 'Memory_size':24, 'ICN':400, 'real_values':True},    # Code Generation
]

System_names_in_plot = ['TPUv5e', 'H100', 'MI300X', 'Gaudi3',
                        'Sys_QA','Sys_Chat', 'Sys_RAG_QA', 'Sys_Summ', 'Sys_Code']

import plotnine as p9
plot_theme = p9.theme(axis_text_x=p9.element_text(size=24),
                axis_text_y=p9.element_text(size=24),
                axis_title_x=p9.element_text(size=24),  # Adjust the size as needed
                axis_title_y=p9.element_text(size=24),
                plot_title=p9.element_text(size=24),  # Adjust the size as needed
                # legend_position='inside',
                legend_position='right',
                # legend_key_spacing_y=20000,
                legend_title=p9.element_text(size=24,margin={'b': 10}),
                legend_text=p9.element_text(size=24),  # Adjust the size as needed
                panel_background=p9.element_rect(fill='white'),
                plot_background=p9.element_rect(fill='white'),
                panel_border=p9.element_rect(color='black', size=1),
                panel_grid_major=p9.element_line(color='black', size=0.5),
                panel_grid_minor=p9.element_line(color='black', size=0.25, linetype='dashed'),
                axis_line=p9.element_line(color='white', size=1),
                strip_text=p9.element_text(size=24),
                figure_size=(20, 6))




import seaborn as sns
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from GenZ.unit import Unit
from IPython.display import display

unit = Unit()

def plot_roofline_background(system, max_x,unit):
    op_intensity = system.flops/system.offchip_mem_bw
    flops = unit.raw_to_unit(system.flops, type='C')
    max_x = max(max_x, op_intensity*2.5)
    turning_points = [[0, 0], [op_intensity, flops], [max_x, flops]]
    turning_points = np.array(turning_points)
    plt.plot(turning_points[:,0], turning_points[:,1], c='grey')

    op_intensity = system.flops/system.onchip_mem_bw
    flops = unit.raw_to_unit(system.flops, type='C')
    turning_points = [[0, 0], [op_intensity, flops], [max_x, flops]]
    turning_points = np.array(turning_points)
    plt.plot(turning_points[:,0], turning_points[:,1], '--', c='grey')

    plt.xlabel('Op Intensity (FLOPs/Byte)')
    plt.ylabel(f'{unit.unit_compute.upper()}')


def dot_roofline(df,system,unit):
    max_x = max(df['Op Intensity'])
    plot_roofline_background(system, max_x,unit)
    for i in range(len(df)):
        op_intensity = df.loc[i, 'Op Intensity']
        thrpt = df.loc[i, 'Throughput (Tflops)']
        plt.scatter(op_intensity, thrpt)

def color_bound_type(value):
    if value == 'Memory':
        color = 'red'
    elif value == 'Compute':
        color = 'green'
    elif value == 'Collective':
        color = 'blue'
    else:
        return
    return 'color: %s' % color

def highlight_max_cycles(s):
    '''
    highlight the maximum in a Series green.
    '''
    is_max = s == s.max()
    return ['background-color: green' if v else '' for v in is_max]

def display_df(df):
    ## Adding % of total time for each operation
    total_cycles = np.sum(df['Cycles'])
    for i in range(len(df)):
        df.loc[i, '% of total time'] = 100*df.loc[i,'Cycles']/total_cycles

    ## reducing display precision
    pd.set_option("display.precision", 2)

    ## Reordering columns
    first_cols = ['Layer Name', 'Op Type','Dimension','Op Intensity', 'Latency (msec)', 'Num ops (MFLOP)','Input_a (MB)','Input_w (MB)','Output (MB)','Total Data (MB)',f'Compute time ({unit.unit_time})',f'Memory time ({unit.unit_time})',f'Communication time ({unit.unit_time})','Bound','C/M ratio','Cycles', '% of total time','Throughput (Tflops)',f'Compute cycle',f'Memory cycle']
    last_cols = [col for col in df.columns if col not in first_cols]
    df = df.loc[:,first_cols+last_cols]

    ## Applying colors and gradients to colmns
    df = df.style.background_gradient(cmap='Blues',axis=0,subset=["Cycles","Total Data (MB)"])\
        .background_gradient(cmap='Spectral_r',axis=1,subset=['Input_a (MB)','Input_w (MB)','Output (MB)'])\
        .background_gradient(cmap='Oranges',axis=0,subset=["Op Intensity"])\
        .map(color_bound_type, subset=['Bound'])\
        .apply(highlight_max_cycles,axis=1,subset=pd.IndexSlice[:, [f'Compute time ({unit.unit_time})',f'Memory time ({unit.unit_time})',f'Communication time ({unit.unit_time})']])    

    pd.set_option('display.float_format', '{:.5f}'.format)
    display(df)
    return df


from ..default_models import ModelConfig, get_all_model_configs

##### Nvidia Models ########
# https://huggingface.co/nvidia/Nemotron-4-340B-Instruct/blob/main/model_config.yaml
nemotron_340b_config = ModelConfig(model='nvidia/Nemotron-4-340B-Instruct',
    hidden_size=18432, num_attention_heads=96,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=73728, num_decoder_layers=96,
    vocab_size=256000, max_model_len=4*1024, hidden_act="silu",
)

# TODO: Addition changes required in Genz as has different layerwise configs
# https://huggingface.co/nvidia/Llama-3_1-Nemotron-51B-Instruct/blob/main/config.json
# nemotron_51b_config = ModelConfig(model='nvidia/Llama-3_1-Nemotron-51B-Instruct',

#     hidden_size=8192, num_attention_heads=64,
#     num_key_value_heads=8, num_ffi = 2,
#     intermediate_size=28672, num_decoder_layers=80,
#     vocab_size=128256, max_model_len=128*1024, hidden_act="silu",
# )


# https://arxiv.org/pdf/2402.16819
nemotron_15b_config = ModelConfig(model='nvidia/Nemotron-4-15B',
    hidden_size=6144, num_attention_heads=48,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=4*6144, num_decoder_layers=32,
    vocab_size=256000, max_model_len=4*1024, hidden_act="relu",
)

# https://huggingface.co/nvidia/Mistral-NeMo-Minitron-8B-Base/blob/main/config.json
minitron_8b_config = ModelConfig(model='nvidia/Mistral-NeMo-Minitron-8B-Base',
    hidden_size=4096, num_attention_heads=32,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=11520, num_decoder_layers=40,
    vocab_size=131072, max_model_len=8*1024, hidden_act="silu",
)

nvidia_models = get_all_model_configs(__name__)
nvidia_models.update({
    # 'nvidia/Llama-3_1-Nemotron-51B-Instruct': nemotron_51b_config,
    'nvidia/Mistral-NeMo-Minitron-8B-Base': minitron_8b_config,
    'nvidia/Nemotron-4-15B': nemotron_15b_config,
    'nvidia/Nemotron-4-340B-Instruct': nemotron_340b_config,

})



from GenZ.Models import ModelConfig, ResidencyInfo, OpType, CollectiveType
from GenZ.parallelism import ParallelismConfig
import warnings
from math import ceil

def ffn_prefill(model_config:ModelConfig, parallelism_config:ParallelismConfig, input_sequence_length:int):

    tp = parallelism_config.tensor_parallel
    ep = parallelism_config.expert_parallel
    sp = parallelism_config.sequence_parallel
    dp = parallelism_config.data_parallel

    D = model_config.hidden_size
    Df = model_config.intermediate_size
    fi = model_config.num_ffi

    E = model_config.num_experts
    K = model_config.expert_top_k
    Df = max(ceil(Df/tp),1)
    moe_layer = (E > 1)

    if E == 1 and ep > 1:
        warnings.warn(f"For dense model, expert parallelism:{ep} will be treated as model parallel")

    assert E % ep == 0, f"Number of experts:{E} must be divisible by expert parallelism:{ep}"

    assert E >= ep, f"Number of experts:{E} must be less than expert parallelism:{ep}"

    layers = []
    if moe_layer:
        router = [["Gate",E, input_sequence_length//sp, D//tp, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]
        layers += router
        if tp > 1:
            router_AR = [["Gate AR",input_sequence_length//sp, E, 1, 1, tp, CollectiveType.AllReduce, OpType.Sync]]
            layers += router_AR
        num_tokens_per_expert = (input_sequence_length//sp) * K // E
        if ep > 1:
            # Total Size=Batch Size×Tokens per Batch×Hidden Dimension×Number of Experts per Token
            dispatch_all2all = [["Dispatch A2A",input_sequence_length//sp, K*D, 1, 1, ep, CollectiveType.All2All, OpType.Sync]]
            layers += dispatch_all2all
        ffup =   [["up+gate",(E//ep)*Df*fi, num_tokens_per_expert, D, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]
        ffdown = [["down",D, num_tokens_per_expert, (E//ep)*Df, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]

        layers += ffup + ffdown
        if ep > 1:
            collect_all2all = [["Collect A2A",input_sequence_length//sp, K*D, 1, 1, ep, CollectiveType.All2All, OpType.Sync]]
            layers += collect_all2all
    else:
        ffup =   [["up+gate",Df*fi, input_sequence_length//sp, D, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]
        ffdown = [["down",D, input_sequence_length//sp, Df, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]
        layers += ffup + ffdown

    if tp > 1:
        sync =          [["FFN AR",input_sequence_length//sp, D, 1, 1, tp, CollectiveType.AllReduce, OpType.Sync]]
    else:
        sync = []

    return layers + sync

def ffn_decode(model_config:ModelConfig, parallelism_config:ParallelismConfig):
    D = model_config.hidden_size
    Df = model_config.intermediate_size
    fi = model_config.num_ffi

    tp = parallelism_config.tensor_parallel
    ep = parallelism_config.expert_parallel
    sp = parallelism_config.sequence_parallel
    dp = parallelism_config.data_parallel

    E = model_config.num_experts
    K = model_config.expert_top_k
    Df = max(ceil(Df/tp),1)
    moe_layer = (E > 1)

    if E == 1 and ep > 1:
        warnings.warn(f"For dense model, expert parallelism:{ep} will be treated as model parallel")

    assert E % ep == 0, f"Number of experts:{E} must be divisible by expert parallelism:{ep}"
    assert E >= ep, f"Number of experts:{E} must be less than expert parallelism:{ep}"


    layers = []
    if moe_layer:
        router = [["Gate",E, 1, D//tp, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]
        layers += router
        if tp > 1:
            router_AR = [["Gate AR",1, E, 1, 1, tp, CollectiveType.AllReduce, OpType.Sync]]
            layers += router_AR
        if ep > 1:
            # Total Size=Batch Size×Tokens per Batch×Hidden Dimension×Number of Experts per Token
            dispatch_all2all = [["Dispatch A2A",1, K*D, 1, 1, ep, CollectiveType.All2All, OpType.Sync]]
            layers += dispatch_all2all

        ## TODO: Define a function to calculate the number of activated experts
        A = K

        npus_activated = max(1,ceil(A/ep))
        ## Understanding load imbalance among experts
        # Lets' say we have 4 experts and 2 experts per token
        # E = 4, k = 2 ,ep = 2, Activated = 2
        # Best case: activated experts are distributed among EP
        #   Then FF: Df*fi*max(1, ceil(A/ep) )
        # Worst case: all activated experts are in the same EP
        #   Then FF: Df*fi*min(A, ceil(E/ep))
        #   E =  16, k = 2, ep = 4, A = 3
        #   Best case: max(1, 3/4) = 1 expert per chip
        #   Worst case: min(4, 16/4) = 4 expert per chip
        #
        #   E =  16, k = 2, ep = 4, A = 5
        #   Best case: max(1, 5//4) = 2 expert per chip
        #   Worst case: min(5, 16//4) = 4 expert per chip

        ## Activated experts are distributed among EP
        ffup =           [["up+gate",npus_activated*Df*fi, 1, D, 1, 1, ResidencyInfo.AC_onchip, OpType.GEMM]]    ## Df is already divided
        ffdown =           [["down",D, 1, npus_activated*Df, 1, 1, ResidencyInfo.AC_onchip, OpType.GEMM]]

        ## These are unused layers but kept just for weights calculation
        ffup_unused =   [["up+gate",(E-A)*Df*fi, 0, D, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]
        ffdown_unused =   [["down",D, 0, (E-A)*Df, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]

        layers += ffup + ffdown + ffup_unused + ffdown_unused
        if ep > 1:
            collect_all2all = [["Collect A2A",1, K*D, 1, 1, ep, CollectiveType.All2All, OpType.Sync]]
            layers += collect_all2all
    else:
        ffup =           [["up+gate",Df*fi, 1, D, 1, 1, ResidencyInfo.AC_onchip, OpType.GEMM]]    ## Df is already divided
        ffdown =           [["down",D, 1, Df, 1, 1, ResidencyInfo.AC_onchip, OpType.GEMM]]
        layers += ffup + ffdown

    if tp > 1:
        sync =          [["FFN AR",1, D, 1, 1, tp, CollectiveType.AllReduce, OpType.Sync]]
    else:
        sync = []

    return layers + sync


from ..default_models import ModelConfig, get_all_model_configs


# https://huggingface.co/tiiuae/falcon-7b-instruct/blob/main/config.json
falcon7b_config = ModelConfig(model='tiiuae/falcon-7b-instruct',
    hidden_size=4544, num_attention_heads=71, num_ffi = 1,
    num_key_value_heads=71, head_dim=64,
    intermediate_size=4544*4, num_decoder_layers=32,
    vocab_size=65024, max_model_len=2*1024, hidden_act="relu",
)

# https://huggingface.co/databricks/dbrx-base/blob/main/config.json
dbrx_config = ModelConfig(model='databricks/dbrx-base',
    hidden_size=6144, num_attention_heads=48,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=10752, num_decoder_layers=40,
    expert_top_k=4, num_experts=16, moe_layer_freq=1,
    max_model_len=32*1024, vocab_size=100352, hidden_act="silu",
)

gpt_4_config = ModelConfig(model='openai/GPT-4',
    hidden_size=84*128, num_attention_heads=84,
    num_key_value_heads=84, num_ffi = 1,
    intermediate_size=4*84*128, num_decoder_layers=128,
    expert_top_k=2, num_experts=16, moe_layer_freq=1,
    vocab_size=128256, max_model_len=128*1024, hidden_act="silu",
)

# https://huggingface.co/xai-org/grok-1/blob/main/RELEASE
grok_1_config = ModelConfig(model='xai-org/grok-1',
    hidden_size=6144, num_attention_heads=48,
    num_key_value_heads=8, num_ffi = 1,
    intermediate_size=8*6144, num_decoder_layers=64,
    expert_top_k=2, num_experts=8, moe_layer_freq=1,
    vocab_size=128*1024, max_model_len=8*1024, hidden_act="gelu",
)

# https://huggingface.co/THUDM/glm-4-9b-chat/blob/main/config.json
glm_9b_config = ModelConfig(model='THUDM/glm-4-9b-chat',
    hidden_size=4096, num_attention_heads=32,
    num_key_value_heads=2, num_ffi = 2,
    intermediate_size=13696, num_decoder_layers=40,
    max_model_len=131072, vocab_size=151552, hidden_act="gelu",
)

# https://huggingface.co/state-spaces/mamba-130m-hf/blob/main/config.json
mamba_130m_config = ModelConfig(model='state-spaces/mamba-130m-hf',
    hidden_size=768, num_attention_heads=1,
    num_key_value_heads=1, num_ffi = 2,
    intermediate_size=1536, num_decoder_layers=24,
    max_model_len=131072, vocab_size=50280, hidden_act="silu",
    mamba_d_state = 16, mamba_dt_rank = 48, mamba_expand = 2, mamba_d_conv=4,
    mamba_layer_freq = 1,
)

# https://huggingface.co/state-spaces/mamba-2.8b-hf/blob/main/config.json
mamba_3b_config = ModelConfig(model='state-spaces/mamba-2.8b-hf',
    hidden_size=2560, num_attention_heads=1,
    num_key_value_heads=1, num_ffi = 2,
    intermediate_size=5120, num_decoder_layers=64,
    max_model_len=131072, vocab_size=50280, hidden_act="silu",
    mamba_d_state = 16, mamba_dt_rank = 160, mamba_expand = 2, mamba_d_conv=4,
    mamba_layer_freq = 1,
)

# https://huggingface.co/tiiuae/falcon-mamba-7b/blob/main/config.json
falcon_mamba_7b_config = ModelConfig(model='tiiuae/falcon-mamba-7b',
    hidden_size=4096, num_attention_heads=1,
    num_key_value_heads=1, num_ffi = 2,
    intermediate_size=8192, num_decoder_layers=64,
    max_model_len=131072, vocab_size=50280, hidden_act="silu",
    mamba_d_state = 16, mamba_dt_rank = 256, mamba_expand = 16, mamba_d_conv=4,
    mamba_layer_freq = 1,
)


super_llm_config = ModelConfig(model='SuperLLM-10T',
    hidden_size=108*128, num_attention_heads=108,
    num_key_value_heads=108, num_ffi = 2,
    intermediate_size=4*108*128, num_decoder_layers=128,
    expert_top_k=4, num_experts=32, moe_layer_freq=1
)



# https://huggingface.co/HuggingFaceH4/zephyr-7b-beta/blob/main/config.json
zephyr_7b_config = ModelConfig(model='HuggingFaceH4/zephyr-7b-beta',
    hidden_size=4096, num_attention_heads=32,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=14336, num_decoder_layers=32,
    vocab_size=32000, max_model_len=32*1024, sliding_window=4096,
    hidden_act="silu")

# https://huggingface.co/deepseek-ai/deepseek-moe-16b-base/blob/main/config.json
deep_seek_moe_16b_config = ModelConfig(model='deepseek-ai/deepseek-moe-16b-base',
    hidden_size=2048, num_attention_heads=16,
    num_key_value_heads=16, num_ffi = 2,
    intermediate_size=10944, num_decoder_layers=28,
    expert_top_k=6, num_experts=64, moe_layer_freq=1,
    moe_intermediate_size=1408,
    n_shared_experts=2, shared_expert_intermediate_size=1408,
    first_k_dense_replace = 1,
    vocab_size=102400, max_model_len=4*1024,  hidden_act="silu",
)
## TODO: account for shared expert, shared account is regular MLP which is always added.
## This has a special case where the first layer is dense and the rest are MoE with shared experts.
## MLP in this case is n_shared_experts*shared_expert_intermediate_size + Activated*moe_intermediate_size

misc_models = get_all_model_configs(__name__)

misc_models.update({
    'glm-9b': glm_9b_config,
    'HuggingFaceH4/zephyr-7b-beta': zephyr_7b_config,
    'THUDM/glm-4-9b-chat': glm_9b_config,
    'databricks/dbrx-base': dbrx_config,
    'dbrx': dbrx_config,
    'falcon7b': falcon7b_config,
    'gpt-4': gpt_4_config,
    'grok-1': grok_1_config,
    'openai/gpt-4': gpt_4_config,
    'state-spaces/mamba-130m-hf': mamba_130m_config,
    'state-spaces/mamba-2.8b-hf': mamba_3b_config,
    'nvidia/mamba2-8b-3t-4k': mamba_3b_config,
    'super_llm': super_llm_config,
    'tiiuae/falcon-7b-instruct': falcon7b_config,
    'tiiuae/falcon-mamba-7b': falcon_mamba_7b_config,
    'xai-org/grok-1': grok_1_config,
})


from ..default_models import ModelConfig, get_all_model_configs


#### Phi3 Models ####

# https://huggingface.co/microsoft/Phi-3-mini-128k-instruct/blob/main/config.json
phi3mini_config = ModelConfig(model='microsoft/Phi-3-mini',
    hidden_size=3072, num_attention_heads=32, num_ffi = 2,
    intermediate_size=8192, num_decoder_layers=32,
    vocab_size=32064, max_model_len=128*1024, hidden_act="silu",
    sliding_window=256*1024
)
# https://huggingface.co/microsoft/Phi-3-small-128k-instruct/blob/main/config.json
phi3small_config = ModelConfig(model='microsoft/Phi-3-small',
    hidden_size=4096, num_attention_heads=32, num_ffi = 2,
    num_key_value_heads=8, head_dim=128,
    intermediate_size=14336, num_decoder_layers=32,
    vocab_size=100352, max_model_len=128*1024, hidden_act="gegelu",
    sliding_window=256*1024
)

# https://huggingface.co/microsoft/Phi-3-medium-4k-instruct/blob/main/config.json
phi3medium_config = ModelConfig(model='microsoft/Phi-3-medium',
    hidden_size=5120, num_attention_heads=40, num_ffi = 2,
    num_key_value_heads=10, head_dim=128,
    intermediate_size=17920, num_decoder_layers=40,
    vocab_size=32064, max_model_len=4*1024, hidden_act="silu",
    sliding_window=2047
)

# https://huggingface.co/microsoft/Phi-3.5-MoE-instruct/blob/main/config.json
phi3moe_config = ModelConfig(model='microsoft/Phi-3.5-MoE',
        hidden_size=4096, num_attention_heads=32,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=6400, num_decoder_layers=32,
    expert_top_k=2, num_experts=16, moe_layer_freq=1,
    vocab_size=32064, max_model_len=128*1024, hidden_act="silu",
    sliding_window=128*1024
)

microsoft_models = get_all_model_configs(__name__)
microsoft_models.update({
    'microsoft/phi3medium': phi3medium_config,
    'microsoft/phi3mini': phi3mini_config,
    'microsoft/phi3small': phi3small_config,
    'microsoft/phi3moe': phi3moe_config,
    'phi3medium': phi3medium_config,
    'phi3mini': phi3mini_config,
    'phi3small': phi3small_config,
})



from GenZ.Models import ModelConfig, ResidencyInfo, OpType, CollectiveType
from GenZ.parallelism import ParallelismConfig
import warnings

def input_embedding(model_config:ModelConfig, parallelism_config:ParallelismConfig, input_sequence_length:int):

    tp = parallelism_config.tensor_parallel * parallelism_config.expert_parallel
    sp = parallelism_config.sequence_parallel

    D = model_config.hidden_size
    V = model_config.vocab_size


    emb =   [["embeddings", D, max(1,input_sequence_length//sp), V//tp, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]

    if tp > 1:
        sync =          [["Emb_AR", max(1,input_sequence_length//sp), D, 1, 1, tp, CollectiveType.AllReduce, OpType.Sync]]
    else:
        sync = []

    return emb + sync

def output_embedding(model_config:ModelConfig, parallelism_config:ParallelismConfig, input_sequence_length:int):

    tp = parallelism_config.tensor_parallel * parallelism_config.expert_parallel
    sp = parallelism_config.sequence_parallel

    D = model_config.hidden_size
    V = model_config.vocab_size


    emb =   [["classifier", V//tp, max(1,input_sequence_length//sp), D, 1, 1, ResidencyInfo.All_offchip, OpType.GEMM]]

    if tp > 1:
        sync =          [["classifier_AG",max(1,input_sequence_length//sp), V, 1, 1, tp, CollectiveType.AllGather, OpType.Sync]]
    else:
        sync = []

    return emb + sync


import pandas as pd
import os
from math import ceil
import numpy as np
from datetime import datetime
from GenZ.parallelism import ParallelismConfig

from GenZ.Models.default_models import ModelConfig, MODEL_DICT

from GenZ.Models.utils import OpType, ResidencyInfo, CollectiveType, parse_einsum_expression
from GenZ.Models.attention import mha_flash_attention_prefill, mha_flash_attention_decode
from GenZ.Models.ffn import ffn_prefill, ffn_decode
from GenZ.Models.mamba import mamba_prefill, mamba_decode
from GenZ.Models.embedding import input_embedding, output_embedding

def get_configs(name) -> ModelConfig:
    name = name.lower()

    if model := MODEL_DICT.get_model(name):
        model_config = model
    else:
        print("ERROR, model name parsed incorrect, please check!!! Model Name:",name)

    return model_config

def save_layers(layers:list, data_path:str, name:str):
    model_path = os.path.join(data_path,"model")
    df = pd.DataFrame(layers, columns=['Name', 'M', 'N', 'D', 'H', 'Z', 'Z', 'T'])
    file_name = name.replace("/", "_") + datetime.now().strftime("%m_%d_%Y_%H_%M_%S") +'.csv'
    if not os.path.exists(model_path):
        os.makedirs(model_path)
    df.to_csv(os.path.join(model_path, file_name),  header=True, index=None)
    return file_name

def repeat_layers(num_repeat:int):
    return [["Repeat", num_repeat, 1, 1, 1, 1, 1, OpType.REPEAT]]

def end_repeat_layers(num_repeat:int):
    return [["End Repeat", num_repeat, 1, 1, 1, 1, 1, OpType.ENDREPEAT]]

DATA_PATH = "/tmp/genz/data/"

def create_inference_moe_prefill_layer(input_sequence_length, name='GPT-2', data_path=DATA_PATH,
                         **args):
    model_config = get_configs(name)

    parallelism_config = ParallelismConfig(
        tensor_parallel=args.get('tensor_parallel',1),
        expert_parallel=args.get('expert_parallel',1),
        sequence_parallel=args.get('sequence_parallel',1),
        data_parallel=args.get('data_parallel',1)
        )

    layers = mha_flash_attention_prefill(model_config, parallelism_config, input_sequence_length) + ffn_prefill(model_config, parallelism_config, input_sequence_length)

    return save_layers(layers=layers, data_path=data_path, name=name+"_prefix_")

def create_inference_moe_decode_layer(input_sequence_length, name='GPT-2', data_path=DATA_PATH,
                         output_gen_tokens=32, **args):

    model_config = get_configs(name)
    parallelism_config = ParallelismConfig(
        tensor_parallel=args.get('tensor_parallel',1),
        expert_parallel=args.get('expert_parallel',1),
        sequence_parallel=args.get('sequence_parallel',1),
        data_parallel=args.get('data_parallel',1)
        )
    layers = mha_flash_attention_decode(model_config, parallelism_config, input_sequence_length, output_gen_tokens) + ffn_decode(model_config, parallelism_config)

    return save_layers(layers=layers, data_path=data_path, name=name+"_decode_")

def create_full_prefill_model(input_sequence_length, name='GPT-2', data_path=DATA_PATH, **args):
    model_config = get_configs(name)
    pipeline_stages = args.get('pipeline_parallel',1)

    parallelism_config = ParallelismConfig(
        tensor_parallel=args.get('tensor_parallel',1),
        expert_parallel=args.get('expert_parallel',1),
        sequence_parallel=args.get('sequence_parallel',1),
        data_parallel=args.get('data_parallel',1),
        )

    def add_layers(layers, num_layers):
        layers += repeat_layers(num_layers)
        layers += mha_flash_attention_prefill(model_config, parallelism_config, input_sequence_length)
        layers += ffn_prefill(model_config, parallelism_config, input_sequence_length)
        layers += end_repeat_layers(num_layers)
        return layers

    full_model = []
    full_model += input_embedding(model_config, parallelism_config, input_sequence_length)
    if pipeline_stages > 1:
        layers_per_stage = ceil(model_config.num_decoder_layers / pipeline_stages)
        layers_last_stage = model_config.num_decoder_layers - layers_per_stage * (pipeline_stages - 1)

        ## For PP stages
        ## First PP-1 stages will have layers_per_stage layers and message pass at the end
        full_model += repeat_layers(pipeline_stages - 1)
        ## Single stage will have layers_per_stage layers
        full_model = add_layers(full_model, layers_per_stage)
        ## Single stage layers end and message pass at the end
        full_model += [["Message Pass", input_sequence_length // args.get('sequence_parallel', 1), model_config.hidden_size, 1, 1, 1, CollectiveType.MessagePass, OpType.Sync]]
        full_model += end_repeat_layers(pipeline_stages - 1)
        ## Last stage will have layers_last_stage layers and no message pass at the end
        full_model = add_layers(full_model, layers_last_stage)
    else:
        full_model = add_layers(full_model, model_config.num_decoder_layers)

    full_model += output_embedding(model_config, parallelism_config, input_sequence_length)
    return save_layers(layers=full_model, data_path=data_path, name=name + "_prefix_")


def create_full_decode_model(input_sequence_length, name='GPT-2', data_path=DATA_PATH, output_gen_tokens=1, **args):

    model_config = get_configs(name)
    pipeline_stages = args.get('pipeline_parallel', 1)

    parallelism_config = ParallelismConfig(
        tensor_parallel=args.get('tensor_parallel', 1),
        expert_parallel=args.get('expert_parallel', 1),
        sequence_parallel=args.get('sequence_parallel', 1),
        data_parallel=args.get('data_parallel', 1),
    )

    def add_layers(layers, num_layers):
        layers += repeat_layers(num_layers)
        layers += mha_flash_attention_decode(model_config, parallelism_config, input_sequence_length, output_gen_tokens)
        layers += ffn_decode(model_config, parallelism_config)
        layers += end_repeat_layers(num_layers)
        return layers

    full_model = []

    if pipeline_stages > 1:
        layers_per_stage = ceil(model_config.num_decoder_layers / pipeline_stages)
        layers_last_stage = model_config.num_decoder_layers - layers_per_stage * (pipeline_stages - 1)

        full_model += repeat_layers(pipeline_stages - 1)
        full_model = add_layers(full_model, layers_per_stage)
        full_model += [["Message Pass", 1, model_config.hidden_size, 1, 1, 1, CollectiveType.MessagePass, OpType.Sync]]
        full_model += end_repeat_layers(pipeline_stages - 1)
        full_model = add_layers(full_model, layers_last_stage)
    else:
        full_model = add_layers(full_model, model_config.num_decoder_layers)
    full_model += output_embedding(model_config, parallelism_config, 1)
    return save_layers(layers=full_model, data_path=data_path, name=name + "_decode_")

def create_inference_mamba_prefix_model(input_sequence_length, name='jamba', data_path=DATA_PATH,
                         **args):

    model_config = get_configs(name)

    parallelism_config = ParallelismConfig(
        tensor_parallel=args.get('tensor_parallel',1),
        expert_parallel=args.get('expert_parallel',1),
        sequence_parallel=args.get('sequence_parallel',1),
        data_parallel=args.get('data_parallel',1)
        )

    layers = mamba_prefill(model_config, parallelism_config, input_sequence_length) + ffn_prefill(model_config, parallelism_config, input_sequence_length)

    return save_layers(layers=layers, data_path=data_path, name=name+"_prefix_")


def create_inference_mamba_decode_model(input_sequence_length, name='jamba', data_path=DATA_PATH,
                         **args):

    model_config = get_configs(name)
    parallelism_config = ParallelismConfig(
        tensor_parallel=args.get('tensor_parallel',1),
        expert_parallel=args.get('expert_parallel',1),
        sequence_parallel=args.get('sequence_parallel',1),
        data_parallel=args.get('data_parallel',1)
        )
    layers = mamba_decode(model_config, parallelism_config, input_sequence_length) + ffn_decode(model_config, parallelism_config)

    return save_layers(layers=layers, data_path=data_path, name=name+"_decode_")



def einsum_test(equation=None, einsum_vars=None):

    if equation is None:
        A = (2, 3, 4)
        B = (2, 4, 5)
        C = (5, 6)
        equation = 'ijk,ikl,lm->ijm'
        einsum_vars = parse_einsum_expression(equation, A, B, C)

    layers = [["test", equation, einsum_vars, 1, 1, 1, ResidencyInfo.All_offchip, OpType.EINSUM]]

    return save_layers(layers=layers, data_path=DATA_PATH, name="einsum_")



def get_AR_time(data, numNodes, system):
    """get_AR_time

    Args:
        data (int): Message size(Bytes) per node to complete all reduce.
        num_AR_nodes (int): Number of nodes among which all-reduce is performed
        system (System object): Object of class System

    Returns:
        time(float): Total time(msec) to complete the All-Reduce
    """
    if data == 0 or numNodes == 1:
        return 0
    ## Ring AR Time = Start Latency + N*Tlink +  2M*(N-1)/(N*BW)
    ## Source:  https://tech.preferred.jp/en/blog/technologies-behind-distributed-deep-learning-allreduce/
    allReduceTime = (5e-6 + 2*(numNodes-1)*system.interchip_link_latency +  2 * (numNodes-1) * (data/numNodes) / system.interchip_link_bw)*1000

    return allReduceTime

def get_AG_time(data, numNodes, system):
    """get_AG_time

    Args:
        data (int): Message size(Bytes) per node to complete all gather.
        num_AG_nodes (int): Number of nodes among which all-gather is performed
        system (System object): Object of class System

    Returns:
        time(float): Total time(msec) to complete the All-Gather
    """
    if data == 0 or numNodes == 1:
        return 0
    ## Ring AG Time = Start Latency + N*Tlink +  2M*(N-1)/(N*BW)
    ## Source:  https://tech.preferred.jp/en/blog/technologies-behind-distributed-deep-learning-allreduce/
    allGatherDuration = (5e-6 + (numNodes-1)*system.interchip_link_latency +  (numNodes-1) * (data/numNodes) / system.interchip_link_bw)*1000

    return allGatherDuration

def get_message_pass_time(data, system):
    """get_message_pass_time

    Args:
        data (int): Message size(Bytes) per node to pass from 1 decide to next.
        system (System object): Object of class System

    Returns:
        time(float): Total time(msec) to pass the Message from 1 node to next
    """
    if data == 0:
        return 0
    msg_pass_time = ((4.2e-6 + (2-1)*system.interchip_link_latency) +  data / system.interchip_link_bw)*1000
    return msg_pass_time


def get_A2A_time(data, numNodes, system):
    """get_A2A_time

    Args:
        data (int): Total message size (Bytes) per node to be exchanged in all-to-all.
        num_A2A_nodes (int): Number of nodes participating in the all-to-all operation.
        system (System object): Object of class System

    Returns:
        time (float): Total time (msec) to complete the All-to-All operation
    """

    ## BWeff = 4B/N if Ring of size N
    ## BWeff = 4B/T if 2D Torus of size TxT

    # M = E/ep * D/tp * F * B * bb

    # A2A time = Start Latency + (N-1) * Tlink + (N-1) * M / BW
    # Where N is the number of nodes, M is the message size per node pair,
    # Tlink is the inter-chip link latency, and BW is the inter-chip memory bandwidth

    if data == 0 or numNodes == 1:
        return 0
    message_size_per_pair = data / numNodes
    A2A_time = (5e-6 + (numNodes - 1) * system.interchip_link_latency +
                (numNodes - 1) * message_size_per_pair / system.interchip_link_bw) * 1000

    return A2A_time


import numpy as np
from GenZ.operator_base import Operator
import ast
from GenZ.Models import OpType, CollectiveType

class FC(Operator):
    def __init__(self, dim, density):
        self.name = dim[0]
        super().__init__(dim=dim[1:], density=density)

    def get_effective_dim_len(self):
        return 3

    def get_tensors(self):
        B, O, I = self.dim[:self.get_effective_dim_len()]
        input_a = (B, I)
        input_w = (O, I)
        output = (B, O)
        return input_a, input_w, output

    def get_gemms(self):
        B, O, I = self.dim[:self.get_effective_dim_len()]
        left = B
        upper = O
        contract = I
        outer = 1
        return left, upper, contract, outer

    def get_num_ops(self):
        B, O, I = self.dim[:self.get_effective_dim_len()]
        return np.prod([B, O, I])

class CONV1D(Operator):
    def __init__(self, dim, density):
        self.name = dim[0]
        super().__init__(dim=dim[1:], density=density)

    def get_effective_dim_len(self):
        return 5

    def get_tensors(self):
        # B -> Batch Size
        # OF -> Number of output channels
        # IF -> Number of input channels
        # N -> Number of elements in the input
        # C -> Number of elements in the kernel
        B, OF, IF, N, C = self.dim[:self.get_effective_dim_len()]
        input_a = (B, N, IF)
        input_w = (OF, 1 , C)
        output = (B, N, OF)
        return input_a, input_w, output

    def get_gemms(self):
        B, OF, IF, N, C = self.dim[:self.get_effective_dim_len()]
        left = B*IF*N
        upper = OF
        contract = C*OF
        outer = 1
        return left, upper, contract, outer


    def get_num_ops(self):
        B, OF, IF, N, C = self.dim[:self.get_effective_dim_len()]
        return  np.prod([B, OF, N, C])

class CONV2D(Operator):
    def __init__(self, dim, density):
        self.name = dim[0]
        super().__init__(dim=dim[1:], density=density)

    def get_effective_dim_len(self):
        return 7

    def get_tensors(self):
        B, K, C, Y, X, R, S = self.dim[:self.get_effective_dim_len()]
        input_a = (B, C, Y, X)
        input_w = (K, C, R, S)
        output = (B, K, Y, X)
        return input_a, input_w, output

    def get_gemms(self):
        B, K, C, Y, X, R, S = self.dim[:self.get_effective_dim_len()]
        left = B*Y*X
        upper = K
        contract = C*R*S
        outer = 1
        return left, upper, contract, outer


    def get_num_ops(self):
        B, K, C, Y, X, R, S = self.dim[:self.get_effective_dim_len()]
        return  np.prod([B, K, C, Y, X, R, S])

class GEMM(Operator):
    def __init__(self, dim, density):
        self.name = dim[0]
        super().__init__(dim=dim[1:], density=density)

    def get_effective_dim_len(self):
        return 4

    def get_dimensions(self):
        return [self.get_tensors()]

    def get_tensors(self):
        B, M, N, K = self.dim[:self.get_effective_dim_len()]
        input_a = (B, K, N)
        input_w = (M, K)
        output = (B, M, N)
        # print(input_a,input_w,output)
        return input_a, input_w, output


    def get_gemms(self):
        B, M, N, K = self.dim[:self.get_effective_dim_len()]
        left = N
        upper = M
        contract = K
        outer = B
        return left, upper, contract, outer


    def get_num_ops(self):
        B, M, N, K = self.dim[:self.get_effective_dim_len()]
        return  np.prod([B, M, N, K])


class Logit(Operator):
    def __init__(self, dim, density):
        self.name = dim[0]
        super().__init__(dim=dim[1:], density=density)

    def get_effective_dim_len(self):
        return 6

    def get_tensors(self):
        ## Refer FLAT paper for more detailed explanation on these parameters
        # B -> Batch Size
        # H -> Number of Heads
        # M -> Seq Len for Q
        # N -> Seq Len for K
        # D -> Split hidden size of key, query and values
        # Hkv -> Number of Heads for K and V.

        B, H, M, N, D, Hkv = self.dim[:self.get_effective_dim_len()]
        assert H % Hkv == 0 , f"H:{H} must be divisible by Hkv:{Hkv} "
        input_a = (B, H, M, D)
        input_w = (B, Hkv, N, D)
        output = (B, H, M, N)
        return input_a, input_w, output

    def get_gemms(self):
        B, H, M, N, D, Hkv = self.dim[:self.get_effective_dim_len()]
        left = M
        upper = N
        contract = D
        outer =B*H
        return left, upper, contract, outer


    def get_num_ops(self):
        B, H, M, N, D, Hkv = self.dim[:self.get_effective_dim_len()]
        return  np.prod([B, H, M, N, D])

class Attend(Operator):
    def __init__(self, dim, density):
        self.name = dim[0]
        super().__init__(dim=dim[1:], density=density)

    def get_effective_dim_len(self):
        return 6

    def get_tensors(self):
        B, H, M, N, D, Hkv = self.dim[:self.get_effective_dim_len()]
        assert H % Hkv == 0 , "H must be divisible by Hkv"
        input_a = (B, H, M, N)
        input_w = (B, Hkv, N, D)
        output = (B, H, M, D)
        return input_a, input_w, output

    def get_gemms(self):
        B, H, M, N, D, Hkv = self.dim[:self.get_effective_dim_len()]
        left = M
        upper = D
        contract = N
        outer = B*H
        return left, upper, contract, outer

    def get_num_ops(self):
        B, H, M, N, D, Hkv = self.dim[:self.get_effective_dim_len()]
        return np.prod([B, H, M, N, D])

class DWCONV(Operator):
    def __init__(self, dim, density):
        self.name = dim[0]
        super().__init__(dim=dim[1:], density=density)

    def get_effective_dim_len(self):
        return 7

    def get_tensors(self):
        B, K, C, Y, X, R, S = self.dim[:self.get_effective_dim_len()]
        input_a = (B, C, Y, X)
        input_w = (C, R, S)
        output = (B, C, Y, X)
        return input_a, input_w, output

    def get_gemms(self):
        B, K, C, Y, X, R, S = self.dim[:self.get_effective_dim_len()]
        left = B*Y*X
        upper = 1
        contract = C*R*S
        outer = 1
        return left, upper, contract, outer

    def get_num_ops(self):
        B, K, C, Y, X, R, S = self.dim[:self.get_effective_dim_len()]
        return  np.prod([B, C, Y, X, R, S])

class Sync(Operator):   ## Just data movement.
    def __init__(self, dim, density):
        self.name = dim[0]
        self.collective_type = dim[-2]
        self.num_collective_nodes = dim[-3]
        super().__init__(dim=dim[1:], density=density)

    def get_effective_dim_len(self):
        return 3

    def get_tensors(self):
        return 0, 0, 0

    def get_dimensions(self):
        if self.num_collective_nodes == 1 and self.collective_type != CollectiveType.MessagePass:
            return None
        else:
            B, M, N = self.dim[:self.get_effective_dim_len()]
            return (B,M,N)

    def communication_data(self):
        if self.num_collective_nodes == 1 and self.collective_type != CollectiveType.MessagePass:
            return 0
        else:
            B, M, N = self.dim[:self.get_effective_dim_len()]
            return B*M*N

    def get_gemms(self):
        left = 0
        upper = 0
        contract = 0
        outer = 0
        return left, upper, contract, outer


    def get_num_ops(self):
        return 0

class Einsum(Operator):
    def __init__(self, dim, density):
        """
        equation: Einstein summation notation string
        dims: Dictionary of tensor dimensions keyed by the corresponding label in the equation
        """
        self.name = dim[0]
        self.batch = dim[1]
        self.equation = dim[2]

        self.dimensions = {k: int(v) if isinstance(v, (int, float)) else v for k, v in ast.literal_eval(dim[3]).items()}
        for k, v in self.dimensions.items():
            if v == 'b':
                self.dimensions[k] = self.batch
            elif isinstance(v, str):
                raise ValueError(f"Unknown dimension {k}:{v} in equation {self.equation}")

        for var in set(''.join(self.equation.split('->')[0].split(','))):
            if var not in set(self.dimensions.keys()):
                raise ValueError(f"Invalid variable {var} in equation {self.equation}")

        super().__init__(dim=dim[1:], density=density)


    def get_tensors(self):
        input_dims = self.equation.split('->')[0]
        input_a = [self.dimensions[label] for label in input_dims.split(',')[0]]
        input_b = [self.dimensions[label] for label in input_dims.split(',')[1]]
        output = [self.dimensions[label] for label in self.equation.split('->')[1]]
        return input_a, input_b, output

    def get_num_ops(self):
        """
        Compute the number of operations needed for the given einsum configuration.
        """
        input_dims = self.equation.split('->')[0]
        dim_labels = set(''.join(input_dims.split(',')))

        # The number of operations is the product of the dimensions involved in the contraction
        num_ops = np.prod([self.dimensions[label] for label in dim_labels])
        return num_ops

class Repeat(Operator):   ## Layer/Model Repetition
    def __init__(self, dim, density):
        self.name = dim[0]
        super().__init__(dim=dim[1:], density=density)

    def get_effective_dim_len(self):
        return 3

    def get_tensors(self):
        return 0, 0, 0
    def get_gemms(self):
        return 0, 0, 0, 0

    def get_dimensions(self):
        _, repeat_count, ID = self.dim[:self.get_effective_dim_len()]
        return repeat_count

    def get_num_ops(self):
        return 0

class EndRepeat(Operator):   ## Layer/Model Repetition
    def __init__(self, dim, density):
        self.name = dim[0]
        super().__init__(dim=dim[1:], density=density)

    def get_effective_dim_len(self):
        return 3

    def get_tensors(self):
        return 0, 0, 0

    def get_dimensions(self):
        _, repeat_count, ID = self.dim[:self.get_effective_dim_len()]
        return repeat_count

    def get_gemms(self):
        return 0, 0, 0, 0

    def get_num_ops(self):
        return 0

class Norm(Operator):          ## Vector Execution Unit
    def __init__(self, dim, density=(1.0, 1.0, 1.0)):
        self.name = dim[0]
        super().__init__(dim, density)
        self.alu_type = 'vxu'

    def get_tensors(self):
        # input and output
        return self.node.inputs[0].shape, self.node.outputs[0].shape

    def get_num_ops(self):
        # macs
        if self.node.operator in ["batch_norm", "instance_norm"]:
            affine = self.node.inputs[1].shape is not None
        elif self.node.operator in ["layer_norm", "group_norm"]:
            affine = self.node.inputs[2].shape is not None
        else:
            raise ValueError(self.node.operator)
        os = self.node.outputs[0].shape
        return np.prod(os) if affine else 0

    def get_gemms(self):
        # TODO might have some problems
        return 0, 0, 0, 1


class Avg(Operator):          ## Vector Execution Unit
    def __init__(self, dim, density=(1.0, 1.0, 1.0)):
        self.name = dim[0]
        super().__init__(dim[1:], density)
        self.alu_type = 'vxu'

    def get_tensors(self):
        # input and output
        return self.node.inputs[0].shape, self.node.outputs[0].shape

    def get_num_ops(self):
        os = self.node.outputs[0].shape
        return np.prod(os)

    def get_gemms(self):
        # TODO might have some problems
        return 0, 0, 0, 1


class Special_Func(Operator):          ## Vector Execution Unit
    def __init__(self, dim, density=(1.0, 1.0, 1.0)):
        self.name = dim[0]
        super().__init__(dim[1:], density)
        self.alu_type = 'vxu'

    def get_tensors(self):
        # input and output
        if self.node.inputs[0].shape is not None:
            return self.node.inputs[0].shape, self.node.outputs[0].shape
        else:
            return self.node.inputs[1].shape, self.node.outputs[0].shape

    def get_num_ops(self):
        os = self.node.outputs[0].shape
        return np.prod(os)

    def get_gemms(self):
        # TODO might have some problems
        return 0, 0, 0, 1




from .utils import ModdelingOutput, get_inference_system, get_offload_system
from GenZ.unit import Unit
from GenZ.operators import *

from GenZ.analyse_model import *
import warnings
from GenZ.collective_times import *
from GenZ.utils.plot_rooflines import *
from GenZ.Models import get_configs, create_full_prefill_model

unit = Unit()

def prefill_moddeling(model = 'BERT', batch_size = 1, input_tokens = 4096,
    system_name = 'A100_40GB_GPU', system_eff=1, bits='bf16', debug= False, model_profilling = False,
    tensor_parallel = 1, pipeline_parallel = 1, expert_parallel = 1,
    collective_strategy='GenZ', network_config=None,
    return_model_df=False, parallelism_heirarchy = "TP{1}_EP{1}_PP{1}",
    model_offload = False):

    if pipeline_parallel > 1:
        ub = max(batch_size // pipeline_parallel, 1)
        num_micro_batches = batch_size // ub
        if batch_size < pipeline_parallel:
            warnings.warn(f"Batch size is divided into micro batches for pipeline parallel, micro batch size:{ub}, consider increasing batch size")
    else:
        ub = batch_size
    ##################################################################################################
    ### System Declaration
    ##################################################################################################

    system = get_inference_system(system_name = system_name, bits = bits, 
                                ceff=system_eff, meff=system_eff, network_config=network_config,
                                collective_strategy=collective_strategy, parallelism_heirarchy=parallelism_heirarchy)
    ##################################################################################################
    ### Model Characterization Calculation
    ##################################################################################################
    model_prefill = create_full_prefill_model(input_sequence_length=input_tokens, name=model,
                                                    tensor_parallel=tensor_parallel,
                                                    pipeline_parallel=pipeline_parallel,
                                                    expert_parallel=expert_parallel)


    model_df = get_model_df(model_prefill, system=system, batch_size = ub, intermediate_on_chip=True , model_characterstics = True)
    summary_table = get_summary_table(model_df, unit, model_characterstics = True)

    model_weights = summary_table[f'Total Weights ({unit.unit_mem})'].values[0]        ## In MB
    kv_cache = summary_table[f'KV Cache ({unit.unit_mem})'].values[0]                  ## In MB

    total_memory_req = model_weights + kv_cache
    num_nodes = pipeline_parallel * tensor_parallel * expert_parallel

    #################################################################################
    ### Offloading calculations
    #################################################################################
    is_offloaded = False
    per_chip_memory = system.get_off_chip_mem_size()   ## MB
    if  per_chip_memory*num_nodes< total_memory_req:
        if model_offload:
            system = get_offload_system(system=system, total_memory_req = total_memory_req , debug=debug)
            warnings.warn(f"Some Parameter offloaded, effective Memory BW:{unit.raw_to_unit(system.offchip_mem_bw, type='BW')} ")
            is_offloaded = True
        elif model_profilling:
            warnings.warn(f"All params would not fit on chip. System Memory Cap:{per_chip_memory/1024} GB , Weights : {model_weights/1024} GB, KV Cache:{kv_cache/1024} ")
        else:
            raise ValueError(f"All params would not fit on chip. System Memory Cap:{per_chip_memory/1024} GB , Weights : {model_weights/1024} GB, KV Cache:{kv_cache/1024}. \n System:{system_name}")

    ## for tensor shareding per layer.
    assert pipeline_parallel >= 1, "Pipeline parallel must be >= 1"
    assert tensor_parallel >= 1, f"Tensor parallel must be >= 1, {tensor_parallel}"
    if model_profilling:
        return model_df, summary_table

    ##################################################################################################
    ### Prefill generation time
    ##################################################################################################
    model_prefill = create_full_prefill_model(input_sequence_length=input_tokens, name=model,
                                                tensor_parallel=tensor_parallel,
                                                pipeline_parallel=pipeline_parallel,
                                                expert_parallel=expert_parallel)
    system.parallelism_heirarchy = parallelism_heirarchy
    model_df = get_model_df(model_prefill, system, unit, ub, intermediate_on_chip=True )
    summary_table = get_summary_table(model_df, unit)
    prefill_latency = summary_table[f'Latency ({unit.unit_time})'].values[0]                 # Latency in millisec
    if return_model_df:
        return model_df, summary_table
    if debug:
        display_df(simplify_df(model_df))
        display(summary_table)

    ##################################################################################################
    ### Final Latency and Thrpt Calculation
    ##################################################################################################

    ## 1000x because the latency is in milli seconds. thrpt is in Token/s
    if pipeline_parallel > 1:
        micro_batch_latency = prefill_latency
        ## If the N micro batches, then the total latency is (N-1)*stage latency + initial_latency
        ## We make the assumption that the pipeline is balanced and the latency is same for all stages
        total_latency = ((num_micro_batches-1) * (prefill_latency / pipeline_parallel)) + micro_batch_latency
        thrpt = 1000 * batch_size / total_latency
    else:
        thrpt = 1000 * batch_size / prefill_latency

    attn_time = summary_table[f'Attn Latency ({unit.unit_time})'].values[0]
    linear_time = summary_table[f'Linear Latency ({unit.unit_time})'].values[0]
    total_communication_delay = summary_table[f'Comm Latency ({unit.unit_time})'].values[0]
    # runtime_breakdown = [linear_time, attn_time, total_communication_delay]
    runtime_breakdown = get_runtime_breakdown(model_df)
    ##################################################################################################
    ### Output Generation
    ##################################################################################################

    return ModdelingOutput(
                        Latency=prefill_latency,
                        Throughput=thrpt,
                        Runtime_breakdown=runtime_breakdown,
                        is_offload=is_offloaded,
                )



import pytest
from GenZ import decode_moddeling, get_model_df, get_configs, System, create_inference_moe_decode_layer, get_AR_time
import os
import pandas as pd
import numpy as np

def test_dense_LLM_decode():
    # Generate the current result
    TPU = System(flops=300, offchip_mem_bw=1200, compute_efficiency=0.8, memory_efficiency=0.8, bits='bf16')
    Model = 'gpt-2'
    Bb = 4
    # Save the current result to a CSV file

    decode_output = decode_moddeling(model = Model, batch_size = 1, input_tokens = 4096, output_tokens=100, Bb=Bb,
                                system_name = TPU, bits='bf16', tensor_parallel = 1, pipeline_parallel = 1, debug=False)


    ref_latency = 5.476083302399999
    ref_throughput = 182.6122695324468
    ref_runtime_breakdown = [2.8991029248, 2.5769803776, 0.0]

    assert np.allclose([decode_output['Latency'], decode_output['Throughput'], decode_output['Runtime_breakdown'][0], decode_output['Runtime_breakdown'][1], decode_output['Runtime_breakdown'][2]],
                        [ref_latency, ref_throughput, ref_runtime_breakdown[0], ref_runtime_breakdown[1], ref_runtime_breakdown[2]])

def test_dense_LLM_decode_with_tensor_parallel():
    # Generate the current result
    TPU = System(flops=300, offchip_mem_bw=1200, compute_efficiency=0.8, memory_efficiency=0.8, bits='bf16',
                interchip_link_bw=50, interchip_link_latency=1)
    Model = 'gpt-2'
    Bb = 4
    # Save the current result to a CSV file
    current_df = get_model_df(model=create_inference_moe_decode_layer(4096, Model, tensor_parallel=4, output_gen_tokens=1024), system=TPU
                            , batch_size=Bb, beam_merge= (Bb > 1), beam_size= Bb)

    ## For GPT-2, the AR message size is 6 KB
    AR_time = get_AR_time(data = 6*2**10, numNodes = 4, system = TPU)

    decode_output = decode_moddeling(model = Model, batch_size = 1, input_tokens = 4096, output_tokens=1024, Bb=Bb, 
                                system_name = TPU, bits='bf16', tensor_parallel = 4, pipeline_parallel = 1, debug=False)

    decode_latency = decode_output['Latency']

    assert decode_latency == (sum(current_df['Latency (msec)']) + 2 * AR_time )* get_configs(Model).num_decoder_layers


import pytest
from GenZ import get_model_df, get_summary_table, System, create_inference_moe_prefill_layer, create_inference_moe_decode_layer

import os
import pandas as pd

def test_dense_LLM_prefill():
    # Delete the current CSV file if it exists
    if os.path.exists('/tmp/current_llama2_7b_prefill_on_TPU.csv'):
        os.remove('/tmp/current_llama2_7b_prefill_on_TPU.csv')

    # Load the golden result
    golden_df = pd.read_csv('./golden/llama2_7b_prefill_on_TPU.csv')

    # Generate the current result
    TPU = System(flops=300, offchip_mem_bw=1200, compute_efficiency=0.8, memory_efficiency=0.8, bits='bf16')

    # Save the current result to a CSV file
    current_df = get_model_df(model=create_inference_moe_prefill_layer(1024, "llama2_7b"), system=TPU)
    current_df.to_csv('/tmp/current_llama2_7b_prefill_on_TPU.csv', index=False)

    # Reload the saved current result
    reloaded_current_df = pd.read_csv('/tmp/current_llama2_7b_prefill_on_TPU.csv')

    # Ensure the reloaded dataframe matches the original current dataframe
    pd.testing.assert_frame_equal(golden_df, reloaded_current_df)

def test_dense_LLM_decode():
    # Delete the current CSV file if it exists
    if os.path.exists('/tmp/current_llama2_7b_decode_on_TPU.csv'):
        os.remove('/tmp/current_llama2_7b_decode_on_TPU.csv')

    # Load the golden result
    golden_df = pd.read_csv('./golden/llama2_7b_decode_on_TPU.csv')

    # Generate the current result
    TPU = System(flops=300, offchip_mem_bw=1200, compute_efficiency=0.8, memory_efficiency=0.8, bits='bf16')

    # Save the current result to a CSV file
    current_df = get_model_df(model=create_inference_moe_decode_layer(1024, "llama2_7b"), system=TPU)
    current_df.to_csv('/tmp/current_llama2_7b_decode_on_TPU.csv', index=False)

    # Reload the saved current result
    reloaded_current_df = pd.read_csv('/tmp/current_llama2_7b_decode_on_TPU.csv')

    # Ensure the reloaded dataframe matches the original current dataframe
    pd.testing.assert_frame_equal(golden_df, reloaded_current_df)

def test_moe_LLM_prefill():
    # Delete the current CSV file if it exists
    if os.path.exists('/tmp/current_mixtral_8x7b_prefill_on_GH200.csv'):
        os.remove('/tmp/current_mixtral_8x7b_prefill_on_GH200.csv')

    # Load the golden result
    golden_df = pd.read_csv('./golden/mixtral_8x7b_prefill_on_GH200.csv')

    # Generate the current result
    GH200 = System(flops=2000, offchip_mem_bw=4900, compute_efficiency=0.8, memory_efficiency=0.8, bits='bf16',
                off_chip_mem_size=144)

    # Save the current result to a CSV file
    current_df = get_model_df(model=create_inference_moe_prefill_layer(1024, "mixtral_8x7b"), system=GH200)
    current_df.to_csv('/tmp/current_mixtral_8x7b_prefill_on_GH200.csv', index=False)

    # Reload the saved current result
    reloaded_current_df = pd.read_csv('/tmp/current_mixtral_8x7b_prefill_on_GH200.csv')

    # Ensure the reloaded dataframe matches the original current dataframe
    pd.testing.assert_frame_equal(golden_df, reloaded_current_df)

def test_moe_LLM_decode():
    # Delete the current CSV file if it exists
    if os.path.exists('/tmp/current_mixtral_8x7b_decode_on_GH200.csv'):
        os.remove('/tmp/current_mixtral_8x7b_decode_on_GH200.csv')

    # Load the golden result
    golden_df = pd.read_csv('./golden/mixtral_8x7b_decode_on_GH200.csv')

    # Generate the current result
    GH200 = System(flops=2000, offchip_mem_bw=4900, compute_efficiency=0.8, memory_efficiency=0.8, bits='bf16',
                off_chip_mem_size=144)

    # Save the current result to a CSV file
    current_df = get_model_df(model=create_inference_moe_decode_layer(1024, "mixtral_8x7b"), system=GH200)
    current_df.to_csv('/tmp/current_mixtral_8x7b_decode_on_GH200.csv', index=False)

    # Reload the saved current result
    reloaded_current_df = pd.read_csv('/tmp/current_mixtral_8x7b_decode_on_GH200.csv')

    # Ensure the reloaded dataframe matches the original current dataframe
    pd.testing.assert_frame_equal(golden_df, reloaded_current_df)


from enum import IntEnum

class OpType(IntEnum):
    FC = 0
    CONV2D = 1
    DWCONV = 2
    GEMM = 3
    Logit = 4
    Attend = 5
    Sync = 6
    Logit_BM_PREFILL = 9
    Attend_BM_PREFILL = 10
    CONV1D = 11
    EINSUM = 12
    REPEAT = 13
    ENDREPEAT = 14
    Norm = 15
    Avg = 16
    Special_Func = 17

class ResidencyInfo(IntEnum):
    All_offchip = 0
    A_onchip = 1
    B_onchip = 2
    C_onchip = 3
    AB_onchip = 4
    AC_onchip = 5
    BC_onchip = 6
    All_onchip = 7

from enum import IntEnum

class CollectiveType(IntEnum):
    AllReduce = 1
    All2All = 2
    AllGather = 3
    ReduceScatter = 4
    MessagePass = 5

class SpecialFuncType(IntEnum):
    gelu = 0
    relu = 1
    softmax = 2
    tanh = 3
    silu = 4
    gelu_pytorch_tanh = 5
    gelu_new = 6
    gegelu = 7

class NormType(IntEnum):
    LayerNorm = 0
    BatchNorm = 1
    InstanceNorm = 2
    GroupNorm = 3


def parse_einsum_expression(expression, *tensors):
    einsum_vars = {}
    input_subscripts, output_subscript = expression.split('->')
    input_subscripts = input_subscripts.split(',')

    for tensor, subscripts in zip(tensors, input_subscripts):
        for dim, subscript in zip(tensor, subscripts):
            if subscript not in einsum_vars:
                einsum_vars[subscript] = dim
            elif einsum_vars[subscript] != dim:
                raise ValueError(f"Dimension mismatch for subscript: {subscript}, Got: {dim}, Expected: {einsum_vars[subscript]}")

    for subscript in output_subscript:
        if subscript not in einsum_vars:
            einsum_vars[subscript] = None

    return einsum_vars